# Time-stamp: <2023-07-28 12:17:28 Tao Liu>

"""Module Description

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

# ------------------------------------
# python modules
# ------------------------------------
import sys
import os
import re
import resource # we turn on memory monitoring inside of logging
from argparse import ArgumentError
from subprocess import Popen, PIPE
from math import log

# ------------------------------------
# MACS3 modules
# ------------------------------------
from MACS3.IO.Parser import BEDParser, ELANDResultParser, ELANDMultiParser, \
    ELANDExportParser, SAMParser, BAMParser, BAMPEParser,\
    BEDPEParser, BowtieParser,  guess_parser

from MACS3.Utilities.Constants import EFFECTIVEGS as efgsize
# ------------------------------------
# constants
# ------------------------------------

import logging
import MACS3.Utilities.Logger
logger = logging.getLogger(__name__)

# ------------------------------------
# Misc functions
# ------------------------------------

logger = logging.getLogger(__name__)

def opt_validate_callpeak ( options ):
    """Validate options from a OptParser object.

    Ret: Validated options object.
    """
    # logging object
    logger.setLevel((4-options.verbose)*10)

    options.error   = logger.critical        # function alias
    options.warn    = logger.warning
    options.debug   = logger.debug
    options.info    = logger.info
    
    # gsize
    try:
        options.gsize = efgsize[options.gsize]
    except:
        try:
            options.gsize = float(options.gsize)
        except:
            logger.error("Error when interpreting --gsize option: %s" % options.gsize)
            logger.error("Available shortcuts of effective genome sizes are %s" % ",".join(list(efgsize.keys())))
            sys.exit(1)

    # format
    options.gzip_flag = False           # if the input is gzip file

    options.format = options.format.upper()
    if options.format == "ELAND":
        options.parser = ELANDResultParser
    elif options.format == "BED":
        options.parser = BEDParser
    elif options.format == "ELANDMULTI":
        options.parser = ELANDMultiParser
    elif options.format == "ELANDEXPORT":
        options.parser = ELANDExportParser
    elif options.format == "SAM":
        options.parser = SAMParser
    elif options.format == "BAM":
        options.parser = BAMParser
        options.gzip_flag = True
    elif options.format == "BAMPE":
        options.parser = BAMPEParser
        options.gzip_flag = True
        options.nomodel = True
    elif options.format == "BEDPE":
        options.parser = BEDPEParser
        options.nomodel = True
    elif options.format == "BOWTIE":
        options.parser = BowtieParser
    elif options.format == "AUTO":
        options.parser = guess_parser
    else:
        logger.error("Format \"%s\" cannot be recognized!" % (options.format))
        sys.exit(1)

    # duplicate reads
    if options.keepduplicates != "auto" and options.keepduplicates != "all":
        if not options.keepduplicates.isdigit():
            logger.error("--keep-dup should be 'auto', 'all' or an integer!")
            sys.exit(1)

    # shiftsize>0
    #if options.shiftsize:               # only if --shiftsize is set, it's true
    #    options.extsize = 2 * options.shiftsize
    #else:                               # if --shiftsize is not set
    #    options.shiftsize = options.extsize / 2
    if options.extsize < 1 :
        logger.error("--extsize must >= 1!")
        sys.exit(1)

    # refine_peaks, call_summits can't be combined with --broad
    #if options.broad and (options.refine_peaks or options.call_summits):
    #    logger.error("--broad can't be combined with --refine-peaks or --call-summits!")
    #    sys.exit(1)

    if options.broad and options.call_summits:
        logger.error("--broad can't be combined with --call-summits!")
        sys.exit(1)

    if options.pvalue:
        # if set, ignore qvalue cutoff
        options.log_qvalue = None
        options.log_pvalue = log(options.pvalue,10)*-1
    else:
        options.log_qvalue = log(options.qvalue,10)*-1
        options.log_pvalue = None
    if options.broad:
        options.log_broadcutoff = log(options.broadcutoff,10)*-1

    # uppercase the format string
    options.format = options.format.upper()

    # d_min is non-negative
    if options.d_min < 0:
        logger.error("Minimum fragment size shouldn't be negative!" % options.d_min)
        sys.exit(1)

    # upper and lower mfold
    options.lmfold = options.mfold[0]
    options.umfold = options.mfold[1]
    if options.lmfold > options.umfold:
        logger.error("Upper limit of mfold should be greater than lower limit!" % options.mfold)
        sys.exit(1)

    # output filenames
    options.peakxls = os.path.join( options.outdir, options.name+"_peaks.xls" )
    options.peakbed = os.path.join( options.outdir, options.name+"_peaks.bed" )
    options.peakNarrowPeak = os.path.join( options.outdir, options.name+"_peaks.narrowPeak" )
    options.peakBroadPeak = os.path.join( options.outdir, options.name+"_peaks.broadPeak" )
    options.peakGappedPeak = os.path.join( options.outdir, options.name+"_peaks.gappedPeak" )
    options.summitbed = os.path.join( options.outdir, options.name+"_summits.bed" )
    options.bdg_treat = os.path.join( options.outdir, options.name+"_treat_pileup.bdg" )
    options.bdg_control= os.path.join( options.outdir, options.name+"_control_lambda.bdg" )
    if options.cutoff_analysis:
        options.cutoff_analysis_file = os.path.join( options.outdir, options.name+"_cutoff_analysis.txt" )
    else:
        options.cutoff_analysis_file = "None"
    #options.negxls  = os.path.join( options.name+"_negative_peaks.xls" )
    #options.diagxls = os.path.join( options.name+"_diag.xls" )
    options.modelR  = os.path.join( options.outdir, options.name+"_model.r" )
    #options.pqtable  = os.path.join( options.outdir, options.name+"_pq_table.txt" )

    options.argtxt = "\n".join((
        "# Command line: %s" % " ".join(sys.argv[1:]),\
        "# ARGUMENTS LIST:",\
        "# name = %s" % (options.name),\
        "# format = %s" % (options.format),\
        "# ChIP-seq file = %s" % (options.tfile),\
        "# control file = %s" % (options.cfile),\
        "# effective genome size = %.2e" % (options.gsize),\
        #"# tag size = %d" % (options.tsize),\
        "# band width = %d" % (options.bw),\
        "# model fold = %s\n" % (options.mfold),\
        ))

    if options.pvalue:
        if options.broad:
            options.argtxt +=  "# pvalue cutoff for narrow/strong regions = %.2e\n" % (options.pvalue)
            options.argtxt +=  "# pvalue cutoff for broad/weak regions = %.2e\n" % (options.broadcutoff)
            options.argtxt +=  "# qvalue will not be calculated and reported as -1 in the final output.\n"
        else:
            options.argtxt +=  "# pvalue cutoff = %.2e\n" % (options.pvalue)
            options.argtxt +=  "# qvalue will not be calculated and reported as -1 in the final output.\n"
    else:
        if options.broad:
            options.argtxt +=  "# qvalue cutoff for narrow/strong regions = %.2e\n" % (options.qvalue)
            options.argtxt +=  "# qvalue cutoff for broad/weak regions = %.2e\n" % (options.broadcutoff)
        else:
            options.argtxt +=  "# qvalue cutoff = %.2e\n" % (options.qvalue)

    if options.maxgap:
        options.argtxt += "# The maximum gap between significant sites = %d\n" % options.maxgap
    else:
        options.argtxt += "# The maximum gap between significant sites is assigned as the read length/tag size.\n"
    if options.minlen:
        options.argtxt += "# The minimum length of peaks = %d\n" % options.minlen
    else:
        options.argtxt += "# The minimum length of peaks is assigned as the predicted fragment length \"d\".\n"

    if options.downsample:
        options.argtxt += "# Larger dataset will be randomly sampled towards smaller dataset.\n"
        if options.seed >= 0:
            options.argtxt += "# Random seed has been set as: %d\n" % options.seed
    else:
        if options.scaleto == "large":
            options.argtxt += "# Smaller dataset will be scaled towards larger dataset.\n"
        else:
            options.argtxt += "# Larger dataset will be scaled towards smaller dataset.\n"

    if options.ratio != 1.0:
        options.argtxt += "# Using a custom scaling factor: %.2e\n" % (options.ratio)

    if options.cfile:
        options.argtxt += "# Range for calculating regional lambda is: %d bps and %d bps\n" % (options.smalllocal,options.largelocal)
    else:
        options.argtxt += "# Range for calculating regional lambda is: %d bps\n" % (options.largelocal)

    if options.broad:
        options.argtxt += "# Broad region calling is on\n"
    else:
        options.argtxt += "# Broad region calling is off\n"

    if options.fecutoff != 1.0:
        options.argtxt += "# Additional cutoff on fold-enrichment is: %.2f\n" % (options.fecutoff)

    if options.format in ["BAMPE", "BEDPE"]:
        # neutralize SHIFT
        options.shift = 0
        options.argtxt += "# Paired-End mode is on\n"
    else:
        options.argtxt += "# Paired-End mode is off\n"

    #if options.refine_peaks:
    #    options.argtxt += "# Refining peak for read balance is on\n"
    if options.call_summits:
        options.argtxt += "# Searching for subpeak summits is on\n"

    if options.do_SPMR and options.store_bdg:
        options.argtxt += "# MACS will save fragment pileup signal per million reads\n"

    return options

def opt_validate_diffpeak ( options ):
    """Validate options from a OptParser object.

    Ret: Validated options object.
    """
    # logging object
    logger.setLevel((4-options.verbose)*10)

    options.error   = logger.critical        # function alias
    options.warn    = logger.warning
    options.debug   = logger.debug
    options.info    = logger.info
    
    # format
    options.gzip_flag = False           # if the input is gzip file

#    options.format = options.format.upper()
    # fox this stuff
#    if True: pass
#    elif options.format == "AUTO":
#        options.parser = guess_parser
#    else:
#        logger.error("Format \"%s\" cannot be recognized!" % (options.format))
#        sys.exit(1)

    if options.peaks_pvalue:
        # if set, ignore qvalue cutoff
        options.peaks_log_qvalue = None
        options.peaks_log_pvalue = log(options.peaks_pvalue,10)*-1
        options.track_score_method = 'p'
    else:
        options.peaks_log_qvalue = log(options.peaks_qvalue,10)*-1
        options.peaks_log_pvalue = None
        options.track_score_method = 'q'

    if options.diff_pvalue:
        # if set, ignore qvalue cutoff
        options.log_qvalue = None
        options.log_pvalue = log(options.diff_pvalue,10)*-1
        options.score_method = 'p'
    else:
        options.log_qvalue = log(options.diff_qvalue,10)*-1
        options.log_pvalue = None
        options.score_method = 'q'

    # output filenames
    options.peakxls = options.name+"_diffpeaks.xls"
    options.peakbed = options.name+"_diffpeaks.bed"
    options.peak1xls = options.name+"_diffpeaks_by_peaks1.xls"
    options.peak2xls = options.name+"_diffpeaks_by_peaks2.xls"
    options.bdglogLR = options.name+"_logLR.bdg"
    options.bdgpvalue = options.name+"_logLR.bdg"
    options.bdglogFC = options.name+"_logLR.bdg"

    options.call_peaks = True
    if not (options.peaks1 == '' or options.peaks2 == ''):
        if options.peaks1 == '':
            raise ArgumentError('peaks1', 'Must specify both peaks1 and peaks2, or neither (to call peaks again)')
        elif options.peaks2 == '':
            raise ArgumentError('peaks2', 'Must specify both peaks1 and peaks2, or neither (to call peaks again)')
        options.call_peaks = False
        options.argtxt = "\n".join((
            "# ARGUMENTS LIST:",\
            "# name = %s" % (options.name),\
#            "# format = %s" % (options.format),\
            "# ChIP-seq file 1 = %s" % (options.t1bdg),\
            "# control file 1 = %s" % (options.c1bdg),\
            "# ChIP-seq file 2 = %s" % (options.t2bdg),\
            "# control file 2 = %s" % (options.c2bdg),\
            "# Peaks, condition 1 = %s" % (options.peaks1),\
            "# Peaks, condition 2 = %s" % (options.peaks2),\
            ""
            ))
    else:
        options.argtxt = "\n".join((
            "# ARGUMENTS LIST:",\
            "# name = %s" % (options.name),\
#            "# format = %s" % (options.format),\
            "# ChIP-seq file 1 = %s" % (options.t1bdg),\
            "# control file 1 = %s" % (options.c1bdg),\
            "# ChIP-seq file 2 = %s" % (options.t2bdg),\
            "# control file 2 = %s" % (options.c2bdg),\
            ""
            ))

        if options.peaks_pvalue:
            options.argtxt +=  "# treat/control -log10(pvalue) cutoff = %.2e\n" % (options.peaks_log_pvalue)
            options.argtxt +=  "# treat/control -log10(qvalue) will not be calculated and reported as -1 in the final output.\n"
        else:
            options.argtxt +=  "# treat/control -log10(qvalue) cutoff = %.2e\n" % (options.peaks_log_qvalue)

    if options.diff_pvalue:
        options.argtxt +=  "# differential pvalue cutoff = %.2e\n" % (options.log_pvalue)
        options.argtxt +=  "# differential qvalue will not be calculated and reported as -1 in the final output.\n"
    else:
        options.argtxt +=  "# differential qvalue cutoff = %.2e\n" % (options.log_qvalue)

    return options

def opt_validate_filterdup ( options ):
    """Validate options from a OptParser object.

    Ret: Validated options object.
    """
    # logging object
    logger.setLevel((4-options.verbose)*10)

    options.error   = logger.critical        # function alias
    options.warn    = logger.warning
    options.debug   = logger.debug
    options.info    = logger.info
    
    # gsize
    try:
        options.gsize = efgsize[options.gsize]
    except:
        try:
            options.gsize = float(options.gsize)
        except:
            logger.error("Error when interpreting --gsize option: %s" % options.gsize)
            logger.error("Available shortcuts of effective genome sizes are %s" % ",".join(list(efgsize.keys())))
            sys.exit(1)

    # format

    options.gzip_flag = False           # if the input is gzip file

    options.format = options.format.upper()
    if options.format == "ELAND":
        options.parser = ELANDResultParser
    elif options.format == "BED":
        options.parser = BEDParser
    elif options.format == "BEDPE":
        options.parser = BEDPEParser
    elif options.format == "ELANDMULTI":
        options.parser = ELANDMultiParser
    elif options.format == "ELANDEXPORT":
        options.parser = ELANDExportParser
    elif options.format == "SAM":
        options.parser = SAMParser
    elif options.format == "BAM":
        options.parser = BAMParser
        options.gzip_flag = True
    elif options.format == "BOWTIE":
        options.parser = BowtieParser
    elif options.format == "BAMPE":
        options.parser = BAMPEParser
        options.gzip_flag = True
    elif options.format == "BEDPE":
        options.parser = BEDPEParser
    elif options.format == "AUTO":
        options.parser = guess_parser
    else:
        logger.error("Format \"%s\" cannot be recognized!" % (options.format))
        sys.exit(1)

    # duplicate reads
    if options.keepduplicates != "auto" and options.keepduplicates != "all":
        if not options.keepduplicates.isdigit():
            logger.error("--keep-dup should be 'auto', 'all' or an integer!")
            sys.exit(1)

    # uppercase the format string
    options.format = options.format.upper()

    return options

def opt_validate_randsample ( options ):
    """Validate options from a OptParser object.

    Ret: Validated options object.
    """
    # logging object
    logger.setLevel((4-options.verbose)*10)

    options.error   = logger.critical        # function alias
    options.warn    = logger.warning
    options.debug   = logger.debug
    options.info    = logger.info
    
    # format

    options.gzip_flag = False           # if the input is gzip file

    options.format = options.format.upper()
    if options.format == "ELAND":
        options.parser = ELANDResultParser
    elif options.format == "BED":
        options.parser = BEDParser
    elif options.format == "ELANDMULTI":
        options.parser = ELANDMultiParser
    elif options.format == "ELANDEXPORT":
        options.parser = ELANDExportParser
    elif options.format == "SAM":
        options.parser = SAMParser
    elif options.format == "BAM":
        options.parser = BAMParser
        options.gzip_flag = True
    elif options.format == "BOWTIE":
        options.parser = BowtieParser
    elif options.format == "BAMPE":
        options.parser = BAMPEParser
        options.gzip_flag = True
    elif options.format == "BEDPE":
        options.parser = BEDPEParser
    elif options.format == "AUTO":
        options.parser = guess_parser
    else:
        logger.error("Format \"%s\" cannot be recognized!" % (options.format))
        sys.exit(1)

    # uppercase the format string
    options.format = options.format.upper()

    # percentage or number
    if options.percentage:
        if options.percentage > 100.0:
            logger.error("Percentage can't be bigger than 100.0. Please check your options and retry!")
            sys.exit(1)
    elif options.number:
        if options.number <= 0:
            logger.error("Number of tags can't be smaller than or equal to 0. Please check your options and retry!")
            sys.exit(1)

    return options

def opt_validate_refinepeak ( options ):
    """Validate options from a OptParser object.

    Ret: Validated options object.
    """
    # logging object
    logger.setLevel((4-options.verbose)*10)

    options.error   = logger.critical        # function alias
    options.warn    = logger.warning
    options.debug   = logger.debug
    options.info    = logger.info
    
    # format

    options.gzip_flag = False           # if the input is gzip file

    options.format = options.format.upper()
    if options.format == "ELAND":
        options.parser = ELANDResultParser
    elif options.format == "BED":
        options.parser = BEDParser
    elif options.format == "ELANDMULTI":
        options.parser = ELANDMultiParser
    elif options.format == "ELANDEXPORT":
        options.parser = ELANDExportParser
    elif options.format == "SAM":
        options.parser = SAMParser
    elif options.format == "BAM":
        options.parser = BAMParser
        options.gzip_flag = True
    elif options.format == "BOWTIE":
        options.parser = BowtieParser
    elif options.format == "AUTO":
        options.parser = guess_parser
    else:
        logger.error("Format \"%s\" cannot be recognized!" % (options.format))
        sys.exit(1)

    # uppercase the format string
    options.format = options.format.upper()

    return options

def opt_validate_predictd ( options ):
    """Validate options from a OptParser object.

    Ret: Validated options object.
    """
    # logging object
    logger.setLevel((4-options.verbose)*10)

    options.error   = logger.critical        # function alias
    options.warn    = logger.warning
    options.debug   = logger.debug
    options.info    = logger.info
    
    # gsize
    try:
        options.gsize = efgsize[options.gsize]
    except:
        try:
            options.gsize = float(options.gsize)
        except:
            logger.error("Error when interpreting --gsize option: %s" % options.gsize)
            logger.error("Available shortcuts of effective genome sizes are %s" % ",".join(list(efgsize.keys())))
            sys.exit(1)

    # format
    options.gzip_flag = False           # if the input is gzip file

    options.format = options.format.upper()
    if options.format == "ELAND":
        options.parser = ELANDResultParser
    elif options.format == "BED":
        options.parser = BEDParser
    elif options.format == "ELANDMULTI":
        options.parser = ELANDMultiParser
    elif options.format == "ELANDEXPORT":
        options.parser = ELANDExportParser
    elif options.format == "SAM":
        options.parser = SAMParser
    elif options.format == "BAM":
        options.parser = BAMParser
        options.gzip_flag = True
    elif options.format == "BAMPE":
        options.parser = BAMPEParser
        options.gzip_flag = True
        options.nomodel = True
    elif options.format == "BEDPE":
        options.parser = BEDPEParser
        options.nomodel = True
    elif options.format == "BOWTIE":
        options.parser = BowtieParser
    elif options.format == "AUTO":
        options.parser = guess_parser
    else:
        logger.error("Format \"%s\" cannot be recognized!" % (options.format))
        sys.exit(1)

    # uppercase the format string
    options.format = options.format.upper()

    # d_min is non-negative
    if options.d_min < 0:
        logger.error("Minimum fragment size shouldn't be negative!" % options.d_min)
        sys.exit(1)

    # upper and lower mfold
    options.lmfold = options.mfold[0]
    options.umfold = options.mfold[1]
    if options.lmfold > options.umfold:
        logger.error("Upper limit of mfold should be greater than lower limit!" % options.mfold)
        sys.exit(1)

    options.modelR  = os.path.join( options.outdir, options.rfile )

    return options


def opt_validate_pileup ( options ):
    """Validate options from a OptParser object.

    Ret: Validated options object.
    """
    # logging object
    logger.setLevel((4-options.verbose)*10)

    options.error   = logger.critical        # function alias
    options.warn    = logger.warning
    options.debug   = logger.debug
    options.info    = logger.info
    
    # format

    options.gzip_flag = False           # if the input is gzip file

    options.format = options.format.upper()
    if options.format == "ELAND":
        options.parser = ELANDResultParser
    elif options.format == "BED":
        options.parser = BEDParser
    elif options.format == "ELANDMULTI":
        options.parser = ELANDMultiParser
    elif options.format == "ELANDEXPORT":
        options.parser = ELANDExportParser
    elif options.format == "SAM":
        options.parser = SAMParser
    elif options.format == "BAM":
        options.parser = BAMParser
        options.gzip_flag = True
    elif options.format == "BOWTIE":
        options.parser = BowtieParser
    elif options.format == "BAMPE":
        options.parser = BAMPEParser
        options.gzip_flag = True
    elif options.format == "BEDPE":
        options.parser = BEDPEParser
    else:
        logger.error("Format \"%s\" cannot be recognized!" % (options.format))
        sys.exit(1)

    # uppercase the format string
    options.format = options.format.upper()

    # extsize
    if options.extsize <= 0 :
        logger.error("--extsize must > 0!")
        sys.exit(1)

    return options

def opt_validate_bdgcmp ( options ):
    """Validate options from a OptParser object.

    Ret: Validated options object.
    """
    # logging object
    logger.setLevel((4-options.verbose)*10)

    options.error   = logger.critical        # function alias
    options.warn    = logger.warning
    options.debug   = logger.debug
    options.info    = logger.info

    # methods should be valid:

    for method in set(options.method):
        if method not in [ 'ppois', 'qpois', 'subtract', 'logFE', 'FE', 'logLR', 'slogLR', 'max' ]:
            logger.error( "Invalid method: %s" % method )
            sys.exit( 1 )

    # # of --ofile must == # of -m

    if options.ofile:
        if len(options.method) != len(options.ofile):
            logger.error("The number and the order of arguments for --ofile must be the same as for -m.")
            sys.exit(1)

    return options


def opt_validate_cmbreps ( options ):
    """Validate options from a OptParser object.

    Ret: Validated options object.
    """
    # logging object
    logger.setLevel((4-options.verbose)*10)

    options.error   = logger.critical        # function alias
    options.warn    = logger.warning
    options.debug   = logger.debug
    options.info    = logger.info

    # methods should be valid:


    if options.method not in [ 'fisher', 'max', 'mean']:
        logger.error( "Invalid method: %s" % options.method )
        sys.exit( 1 )

    if len( options.ifile ) < 2:
        logger.error("Combining replicates needs at least two replicates!")
        sys.exit( 1 )

    # # of -i must == # of -w

    # if not options.weights:
    #     options.weights = [ 1.0 ] * len( options.ifile )

    # if len( options.ifile ) != len( options.weights ):
    #     logger.error("Must provide same number of weights as number of input files.")
    #     sys.exit( 1 )

    # if options.method == "fisher" and len( options.ifile ) > 3:
    #     logger.error("NOT IMPLEMENTED! Can't combine more than 3 replicates using Fisher's method.")
    #     sys.exit( 1 )

    return options


def opt_validate_bdgopt ( options ):
    """Validate options from a OptParser object.

    Ret: Validated options object.
    """
    # logging object
    logger.setLevel((4-options.verbose)*10)

    options.error   = logger.critical        # function alias
    options.warn    = logger.warning
    options.debug   = logger.debug
    options.info    = logger.info

    # methods should be valid:

    if options.method.lower() not in [ 'multiply', 'add', 'p2q', 'max', 'min']:
        logger.error( "Invalid method: %s" % options.method )
        sys.exit( 1 )

    if options.method.lower() in [ 'multiply', 'add' ] and not options.extraparam:
        logger.error( "Need EXTRAPARAM for method multiply or add!")
        sys.exit( 1 )

    return options

def opt_validate_callvar ( options ):
    """Validate options from a OptParser object.

    Ret: Validated options object.
    """
    # logging object
    logger.setLevel((4-options.verbose)*10)

    options.error   = logger.critical        # function alias
    options.warn    = logger.warning
    options.debug   = logger.debug
    options.info    = logger.info

    # methods should be valid:

    if options.np <= 0:
        options.np = 1
    return options


def opt_validate_hmmratac ( options ):
    """Validate options from a OptParser object.

    Ret: Validated options object.
    """
    # logging object
    logger.setLevel((4-options.verbose)*10)

    options.error   = logger.critical        # function alias
    options.warn    = logger.warning
    options.debug   = logger.debug
    options.info    = logger.info
   
    # input options.argtxt for hmmratac
    options.argtxt = "# Command line: %s\n" % " ".join(sys.argv[1:])
    #        "# ARGUMENTS LIST:",\
    #        "# outfile = %s" % (options.ofile),\
    #        "# input file = %s\n" % (options.bam_file),\
    # ... add additional

    # Output options
    #if options.store_bdg:
    #    options.argtxt += "# HMMRATAC will report whole genome bedgraph of all state annotations. \n"
    
    #if options.store_bgscore:
    #    options.argtxt += "# HMMRATAC score will be added to each state annotation in bedgraph. \n"

    #if options.store_peaks:
    #    options.argtxt += "# Peaks not reported in bed format\n"

    #if options.print_exclude:
    #    options.print_exclude = os.path.join(options.outdir, options.ofile+"Output_exclude.bed")
    #else:
    #    options.print_exclude = "None"
    
    #if options.print_train:
    #    options.print_train = os.path.join(options.outdir, options.ofile+"Output_training.bed")
    #else:
    #    options.print_train = "None"


    # EM
    # em_skip
    if options.em_skip:
        options.argtxt += "# EM training not performed on fragment distribution. \n"
    # em_means non-negative
    if sum( [ x < 0 for x in options.em_means ] ):
        logger.error(" --means should not be negative! ")
        sys.exit( 1 )
    # em_stddev non-negative
    if sum( [ x < 0 for x in options.em_stddevs ] ):
        logger.error(" --stddev should not be negative! ")
        sys.exit( 1 )

    # min_frag_p between 0 and 1
    if options.min_frag_p <=0 or options.min_frag_p >= 1:
        logger.error(" --min-frag-p should be larger than 0 and smaller than 1! ")
        sys.exit( 1 )

    # HMM
    # hmm_states non-negative int, warn if not k=3
    #if options.hmm_states <=0:
    #    logger.error(" -s, --states must be an integer >= 0.")
    #    sys.exit( 1 )
    #elif options.hmm_states != 3 and options.hmm_states > 0 and options.store_peaks == False:
    #    logger.warn(" If -s, --states not k=3, recommend NOT calling peaks, use bedgraph.")

    # hmm_binsize > 0
    if options.hmm_binsize <=0:
        logger.error(" --binsize must be larger than 0.")
        sys.exit( 1 )

    # hmm_lower less than hmm_upper, non-negative 
    if options.hmm_lower <0:
        logger.error(" -l, --lower should not be negative! ")
        sys.exit( 1 )
    if options.hmm_upper <0:
        logger.error(" -u, --upper should not be negative! ")
        sys.exit( 1 )
    if options.hmm_lower > options.hmm_upper:
        logger.error("Upper limit of fold change range should be greater than lower limit!" % options.mfold)
        sys.exit(1)
    
    # hmm_maxTrain non-negative
    if options.hmm_maxTrain <= 0:
        logger.error(" --maxTrain should be larger than 0!")
        sys.exit( 1 )
    
    # hmm_training_regions
    if options.hmm_training_regions:
       options.argtxt += "# Using -t, --training input to train HMM instead of using fold change settings to select. \n"
    
    # hmm_zscore non-negative
    #if options.hmm_zscore <0:
    #    logger.error(" -z, --zscore should not be negative!")
    #    sys.exit( 1 )
    
    # hmm_randomSeed
    if options.hmm_randomSeed:
        options.argtxt += "# Random seed selected as: %d\n" % options.hmm_randomSeed
    
    # hmm_window non-negative
    #if options.hmm_window <0:
    #    logger.error(" --window should not be negative! ")
    #    sys.exit( 1 )

    # hmm_file
    #if options.hmm_file:
    #    options.argtxt += "# HMM training will be skipped, --model input used instead. \n"

    # hmm_modelonly
    if options.hmm_modelonly:
       options.argtxt += "# Program will stop after generating model, which can be later applied with '--model'. \n"


    # Peak Calling
    if options.prescan_cutoff <= 1:
        logger.error(" In order to use -c or --prescan-cutoff, the cutoff must be larger than 1.")
        sys.exit( 1 )
    
    if options.openregion_minlen < 0: # and options.store_peaks == True:
        logger.error(" In order to use --minlen, the length should not be negative.")
        sys.exit( 1 )

    #if options.call_score.lower() not in [ 'max', 'ave', 'med', 'fc', 'zscore', 'all']:
    #    logger.error( " Invalid method: %s" % options.call_score )
    #    sys.exit( 1 )

    # call_threshold non-negative
    #if options.call_threshold <0:
    #    logger.error(" --threshold should not be negative! ")
    #    sys.exit( 1 )
    

    # Misc
    # misc_blacklist 
    #if options.misc_keep_duplicates:
    #    options.argtxt += "# Duplicate reads from analysis will be stored. \n"

    # misc_trim non-negative
    #if options.misc_trim <0:
    #    logger.error(" --trim should not be negative! ")
    #    sys.exit( 1 )

    # np # should this be mp? non-negative
    #if options.np <0:
    #    logger.error(" -m, --multiple-processing should not be negative! ")
    #    sys.exit( 1 )
    
    # min_map_quality non-negative
    #if options.min_map_quality <0:
    #    logger.error(" -q, --minmapq should not be negative! ")
    #    sys.exit( 1 )

    
    return options





# Time-stamp: <2021-02-24 17:31:59 Tao Liu>

"""Description: macs call

Copyright (c) 2017 Tao Liu <tliu4@buffalo.edu>

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file COPYING included
with the distribution).

@status: release candidate
@version: $Id$
@author:  Tao Liu
@contact: tliu4@buffalo.edu
"""

# ------------------------------------
# python modules
# ------------------------------------

import logging
import datetime
import sys
from functools import partial
import multiprocessing as mp

from time import time
from math import ceil

# ------------------------------------
# own python modules
# ------------------------------------
from MACS3.Utilities.Constants import *
from MACS3.IO.PeakIO import PeakIO
from MACS3.IO.BAM import BAMParser
from MACS3.Signal.RACollection import RACollection
from MACS3.Signal.PeakVariants import PeakVariants


VCFHEADER_0="""##fileformat=VCFv4.1
##fileDate=%s
##source=MACS_V%s
##Program_Args=%s
##INFO=<ID=M,Number=.,Type=String,Description="MACS Model with minimum BIC value">
##INFO=<ID=MT,Number=.,Type=String,Description="Mutation type: SNV/Insertion/Deletion">
##INFO=<ID=DPT,Number=1,Type=Integer,Description="Depth Treatment: Read depth in ChIP-seq data">
##INFO=<ID=DPC,Number=1,Type=Integer,Description="Depth Control: Read depth in control data">
##INFO=<ID=DP1T,Number=.,Type=String,Description="Read depth of top1 allele in ChIP-seq data">
##INFO=<ID=DP2T,Number=.,Type=String,Description="Read depth of top2 allele in ChIP-seq data">
##INFO=<ID=DP1C,Number=.,Type=String,Description="Read depth of top1 allele in control data">
##INFO=<ID=DP2C,Number=.,Type=String,Description="Read depth of top2 allele in control data">
##INFO=<ID=lnLHOMOMAJOR,Number=1,Type=Float,Description="Log(e) scaled genotype likelihoods of homozygous with major allele model">
##INFO=<ID=lnLHOMOMINOR,Number=1,Type=Float,Description="Log(e) scaled genotype likelihoods of homozygous with minor allele model">
##INFO=<ID=lnLHETERNOAS,Number=1,Type=Float,Description="Log(e) scaled genotype likelihoods of heterozygous with no allele-specific model">
##INFO=<ID=lnLHETERAS,Number=1,Type=Float,Description="Log(e) scaled genotype likelihoods of heterozygous with allele-specific model">
##INFO=<ID=BICHOMOMAJOR,Number=1,Type=Float,Description="BIC value of homozygous with major allele model">
##INFO=<ID=BICHOMOMINOR,Number=1,Type=Float,Description="BIC value of homozygous with minor allele model">
##INFO=<ID=BICHETERNOAS,Number=1,Type=Float,Description="BIC value of heterozygous with no allele-specific model">
##INFO=<ID=BICHETERAS,Number=1,Type=Float,Description="BIC value of heterozygous with allele-specific model">
##INFO=<ID=GQHOMO,Number=1,Type=Integer,Description="Genotype quality of homozygous with major allele model">
##INFO=<ID=GQHETERNOAS,Number=1,Type=Integer,Description="Genotype quality of heterozygous with no allele-specific model">
##INFO=<ID=GQHETERAS,Number=1,Type=Integer,Description="Genotype quality of heterozygous with allele-specific model">
##INFO=<ID=GQHETERASsig,Number=1,Type=Integer,Description="Genotype quality of allele-specific significance compared with no allele-specific model">
##INFO=<ID=AR,Number=1,Type=Float,Description="Estimated allele ratio of heterozygous with allele-specific model">
##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
##FORMAT=<ID=DP,Number=1,Type=Integer,Description="read depth after filtering">
##FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype quality score">"""

VCFHEADER="""##fileformat=VCFv4.1
##fileDate=%s
##source=MACS_V%s
##Program_Args=%s
##INFO=<ID=M,Number=.,Type=String,Description="MACS Model with minimum BIC value">
##INFO=<ID=MT,Number=.,Type=String,Description="Mutation type: SNV/Insertion/Deletion">
##INFO=<ID=DPT,Number=1,Type=Integer,Description="Depth Treatment: Read depth in ChIP-seq data">
##INFO=<ID=DPC,Number=1,Type=Integer,Description="Depth Control: Read depth in control data">
##INFO=<ID=DP1T,Number=.,Type=String,Description="Read depth of top1 allele in ChIP-seq data">
##INFO=<ID=DP2T,Number=.,Type=String,Description="Read depth of top2 allele in ChIP-seq data">
##INFO=<ID=DP1C,Number=.,Type=String,Description="Read depth of top1 allele in control data">
##INFO=<ID=DP2C,Number=.,Type=String,Description="Read depth of top2 allele in control data">
##INFO=<ID=DBIC,Number=.,Type=Float,Description="Difference of BIC of selected model vs second best alternative model">
##INFO=<ID=BICHOMOMAJOR,Number=1,Type=Integer,Description="BIC of homozygous with major allele model">
##INFO=<ID=BICHOMOMINOR,Number=1,Type=Integer,Description="BIC of homozygous with minor allele model">
##INFO=<ID=BICHETERNOAS,Number=1,Type=Integer,Description="BIC of heterozygous with no allele-specific model">
##INFO=<ID=BICHETERAS,Number=1,Type=Integer,Description="BIC of heterozygous with allele-specific model">
##INFO=<ID=AR,Number=1,Type=Float,Description="Estimated allele ratio of heterozygous with allele-specific model">
##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read depth after filtering bad reads">
##FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype Quality score">
##FORMAT=<ID=PL,Number=3,Type=Integer,Description="Normalized, Phred-scaled genotype likelihoods for 00, 01, 11 genotype">"""

# ------------------------------------
# Main function
# ------------------------------------

def check_names(treat, control, error_stream):
    """check common chromosome names"""
    tchrnames = set(treat.get_chr_names())
    cchrnames = set(control.get_chr_names())
    commonnames = tchrnames.intersection(cchrnames)
    if len(commonnames)==0:
        error_stream("No common chromosome names can be found from treatment and control! Check your input files! MACS will quit...")
        error_stream("Chromosome names in treatment: %s" % ",".join(sorted(tchrnames)))
        error_stream("Chromosome names in control: %s" % ",".join(sorted(cchrnames)))
        sys.exit()


def run( args ):
    """The Main function/pipeline for MACS

    """
    peakbedfile = args.peakbed
    peaktfile = args.tfile
    peakcfile = args.cfile
    top2allelesminr = args.top2allelesMinRatio
    min_altallele_count = args.altalleleMinCount
    max_allowed_ar = args.maxAR
    NP = args.np
    if NP<=0:
        NP = 1
    min_homo_GQ = args.GQCutoffHomo
    min_heter_GQ = args.GQCutoffHetero
    minQ = args.Q
    maxDuplicate = args.maxDuplicate

    # parameter for assembly
    fermiMinOverlap = args.fermiMinOverlap
    fermi = args.fermi
    
    peakio = open( peakbedfile )
    peaks = PeakIO()
    i = 0
    for l in peakio:
        fs = l.rstrip().split()
        i += 1
        peaks.add( fs[0].encode(), int(fs[1]), int(fs[2]), name=b"%d" % i )
    peaks.sort()

    chrs = peaks.get_chr_names()

    tbam = BAMParser( peaktfile )
    if peakcfile:
        cbam = BAMParser( peakcfile )
        assert tbam.get_chromosomes()[0] in cbam.get_chromosomes() or cbam.get_chromosomes()[0] in tbam.get_chromosomes(), Exception("It seems Treatment and Control BAM use different naming for chromosomes! Check headers of both files.")
        #assert tbam.get_chromosomes() == cbam.get_chromosomes(), Exception("Treatment and Control BAM files have different orders of sorted chromosomes! Please check BAM Headers and re-sort BAM files.")
    else:
        cbam = None


    ra_collections = []

    # prepare and write header of output file (.vcf)
    ovcf = open(args.ofile, "w")
    tmpcmdstr = " --fermi "+ fermi+ " --fermi-overlap "+str(fermiMinOverlap)
    ovcf.write ( VCFHEADER % (datetime.date.today().strftime("%Y%m%d"), MACS_VERSION, " ".join(sys.argv[1:] + ["-Q", str(minQ), "-D", str(maxDuplicate), "--max-ar", str(max_allowed_ar), "--top2alleles-mratio", str(top2allelesminr), "--top2allele-count", str(min_altallele_count), "-g", str(min_heter_GQ), "-G", str(min_homo_GQ), tmpcmdstr]) ) + "\n" )
    for (chrom, chrlength) in tbam.get_rlengths().items():
        ovcf.write( "##contig=<ID=%s,length=%d,assembly=NA>\n" % ( chrom.decode(), chrlength ) )
    ovcf.write ( "\t".join( ("#CHROM","POS","ID","REF","ALT","QUAL","FILTER","INFO","FORMAT","SAMPLE") ) + "\n" )

    # to get time
    t_total = 0
    t_prepare_ra = 0
    t_assemble = 0
    #t_call_top2alleles = 0
    #t_call_lnL = 0
    t_call_variants = 0
    t_call_GT = 0
    #t_call_to_vcf = 0
    t_total_0 = time()

    for chrom in tbam.get_chromosomes():
        peaks_chr = peaks.get_data_from_chrom( chrom )
        for peak in peaks_chr:
            # note, when we extract reads from BAM within a peak
            # region, we assume BAM should be sorted and the BAM
            # should be generated from "samtools view -L" process.

            # print ( "---begin of peak---")
            print ( "Peak:", chrom.decode(), peak["start"], peak["end"])

            flag_todo_lassembly = False

            t0 = time()
            try:
                if cbam:
                    ra_collection = RACollection( chrom, peak, tbam.get_reads_in_region( chrom, peak["start"], peak["end"], maxDuplicate=maxDuplicate ), cbam.get_reads_in_region( chrom, peak["start"], peak["end"], maxDuplicate=maxDuplicate) )
                else:
                    ra_collection = RACollection( chrom, peak, tbam.get_reads_in_region( chrom, peak["start"], peak["end"], maxDuplicate=maxDuplicate ) )
            except:
                print ("No reads found in peak: ", chrom.decode(), peak["start"], peak["end"], ". Skipped!")
                # while there is no reads in peak region, simply skip it.
                continue
            
            ra_collection.remove_outliers( percent = 5 )
            t_prepare_ra += time() - t0

            # print ( "Reads in Peak:")
            # print ( ra_collection.get_FASTQ().decode() )

            s = ra_collection["peak_refseq"]

            peak_variants = PeakVariants( chrom.decode(), peak["start"], peak["end"], s )
            

            if fermi == "auto" or fermi == "off":
                # first pass to call variant w/o assembly
                # multiprocessing the following part
                t_call_variants_0 = time()
                print ( " Call varants w/o assembly")

                # -- now make multi processes
                # divide right-left into NP parts
                window_size = ceil( ( ra_collection["right"] - ra_collection["left"] ) / NP )

                P = mp.Pool( NP )
                # this partial function will only be used in multiprocessing
                p_call_variants_at_range =  partial(call_variants_at_range, s=s, collection=ra_collection, top2allelesminr=top2allelesminr, max_allowed_ar = max_allowed_ar, min_altallele_count = min_altallele_count, min_homo_GQ = min_homo_GQ, min_heter_GQ = min_heter_GQ, minQ=minQ)

                ranges = []
                for i in range( NP ):
                    l = i * window_size + ra_collection["left"]
                    r = min( (i + 1) * window_size + ra_collection["left"], ra_collection["right"] )
                    ranges.append( (l, r) )

                mapresults = P.map_async( p_call_variants_at_range, ranges )
                P.close()
                P.join()
                results = mapresults.get(timeout=window_size*300)
                for i in range( NP ):
                    for result in results[ i ]:
                        peak_variants.add_variant( result[0], result[1] )

                t_call_variants += time() - t_call_variants_0

            # Next, check if we should do local assembly
            if ( fermi == "auto" and ( peak_variants.has_indel() or peak_variants.has_refer_biased_01() ) ) or fermi == "on":
                #print( peak_variants.has_indel() )
                #print( peak_variants.has_refer_biased_01() )
                    
                # invoke fermi to assemble local sequence and filter out those can not be mapped to unitigs.
                print ( " Assemble using fermi-lite")
                unitig_collection = ra_collection.build_unitig_collection( fermiMinOverlap )
                if unitig_collection == -1:
                    print(" Too many mismatches found while assembling the sequence, we will skip this region entirely!")
                    continue
                elif unitig_collection == 0:
                    print ( "  Failed to assemble unitigs, fall back to previous results" )
                    if peak_variants.n_variants() > 0:
                        peak_variants.fix_indels()
                        ovcf.write( peak_variants.toVCF() )
                        continue
                # uncomment the following to print those assembled unitigs and their alignments to reference genome
                #for u in unitig_collection["URAs_list"]:
                #    print( u["seq"].decode(), u["lpos"], u["rpos"], u["count"] )
                #    print( "a",u["unitig_aln"].decode() )
                #    print( "r",u["reference_aln"].decode() )
            else:
                # if we do not assemble, write results now
                if peak_variants.n_variants() > 0:
                    peak_variants.fix_indels()                    
                    ovcf.write( peak_variants.toVCF() )
                    continue

            # reach here only if we need l assembly and the assembly returns result

            # If a peak has no indel but with refer_biased_01, we
            # revisit all refer_biased_01 now. We do not use
            # multiprocessing here for simplicity since there won't be
            # too many in a peak region.
            if ( fermi == "auto" and ( not peak_variants.has_indel() ) and peak_variants.has_refer_biased_01()  ):
                pos_tobe_revisit = peak_variants.get_refer_biased_01s()
                for i in pos_tobe_revisit:
                    ref_nt = chr(s[ i-ra_collection["left"] ] ).encode()
                    if ref_nt == b'N':
                        peak_variants.remove_variant( i )
                        continue
                    PRI = unitig_collection.get_PosReadsInfo_ref_pos ( i, ref_nt, Q=minQ )
                    if PRI.raw_read_depth( opt="T" ) == 0: # skip if coverage is 0
                        peak_variants.remove_variant( i )                        
                        continue
                    PRI.update_top_alleles( top2allelesminr, min_altallele_count, max_allowed_ar )
                    PRI.call_GT( max_allowed_ar )
                    PRI.apply_GQ_cutoff(min_homo_GQ, min_heter_GQ)
                    if not PRI.filterflag():
                        peak_variants.replace_variant( i, PRI.toVariant() )
                    else:
                        peak_variants.remove_variant( i )
                if peak_variants.n_variants() > 0:
                    peak_variants.fix_indels()
                    ovcf.write( peak_variants.toVCF() )
                    continue

            # in this case, we call variants at every locations in the peak based on local assembly.
            if ( fermi == "on" or ( fermi == "auto" and peak_variants.has_indel() ) ):
                peak_variants = PeakVariants( chrom.decode(), peak["start"], peak["end"], s ) #reset

                # --- make multi processes
                # divide right-left into NP parts
                window_size = ceil( ( ra_collection["right"] - ra_collection["left"] ) / NP )
                P = mp.Pool( NP )
                # this partial function will only be used in multiprocessing
                p_call_variants_at_range =  partial(call_variants_at_range, s=s, collection=unitig_collection, top2allelesminr=top2allelesminr, max_allowed_ar = max_allowed_ar, min_altallele_count = min_altallele_count, min_homo_GQ = min_homo_GQ, min_heter_GQ = min_heter_GQ, minQ=minQ)
                ranges = []
                for i in range( NP ):
                    l = i * window_size + ra_collection["left"]
                    r = min( (i + 1) * window_size + ra_collection["left"], ra_collection["right"] )
                    ranges.append( (l, r) )

                mapresults = P.map_async( p_call_variants_at_range, ranges )
                P.close()
                P.join()
                results = mapresults.get(timeout=window_size*300)                
                for i in range( NP ):
                    for result in results[ i ]:
                        peak_variants.add_variant( result[0], result[1] )
                                
            if peak_variants.n_variants() > 0:
                peak_variants.fix_indels()
                ovcf.write( peak_variants.toVCF() )

    #print ("time to retrieve read alignment information from BAM:",t_prepare_ra,"(",round( 100 * t_prepare_ra/t_total, 2),"% )")
    return

def call_variants_at_range ( lr, s, collection, top2allelesminr, max_allowed_ar, min_altallele_count, min_homo_GQ, min_heter_GQ, minQ ):
#def call_variants_at_range ( lr, chrom, s, collection, top2allelesminr, max_allowed_ar, min_homo_GQ, min_heter_GQ ):
    result = []
    for i in range( lr[ 0 ], lr[ 1 ] ):
        ref_nt = chr(s[ i-collection["left"] ] ).encode()
        if ref_nt == b'N':
            continue

        PRI = collection.get_PosReadsInfo_ref_pos ( i, ref_nt, Q=minQ )
        if PRI.raw_read_depth( opt="T" ) == 0: # skip if coverage is 0
            continue

        PRI.update_top_alleles( top2allelesminr, min_altallele_count, max_allowed_ar )
        if not PRI.filterflag():
            #PRI.update_top_alleles( top2allelesminr )
            PRI.call_GT( max_allowed_ar )
            PRI.apply_GQ_cutoff(min_homo_GQ, min_heter_GQ)
        if not PRI.filterflag():
            result.append( ( i, PRI.toVariant() ) )
    return result





# cython: language_level=3
# cython: profile=True
# Time-stamp: <2020-12-04 18:41:28 Tao Liu>

"""Module for SAPPER BAMParser class

Copyright (c) 2017 Tao Liu <tliu4@buffalo.edu>

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file COPYING included
with the distribution).

@status:  experimental
@version: $Revision$
@author:  Tao Liu
@contact: tliu4@buffalo.edu
"""

# ------------------------------------
# python modules
# ------------------------------------
from cpython cimport bool

cimport cython

import numpy as np
cimport numpy as np

ctypedef np.float32_t float32_t
ctypedef np.int32_t int32_t

#from libc.math cimport log10, log, exp, M_LN10 #,fabs,log1p
#from libc.math cimport M_LN10
from math import log1p, exp, log

LN10 = 2.3025850929940458
LN10_tenth = 0.23025850929940458

@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
cpdef tuple CalModel_Homo( np.ndarray[int32_t, ndim=1] top1_bq_T, np.ndarray[int32_t, ndim=1] top1_bq_C, np.ndarray[int32_t, ndim=1] top2_bq_T, np.ndarray[int32_t, ndim=1] top2_bq_C):
    """Return (lnL, BIC).

    """
    cdef:
        int i
        double lnL, BIC

    lnL=0
    # Phred score is Phred = -10log_{10} E, where E is the error rate.
    # to get the 1-E: 1-E = 1-exp( Phred/-10*M_LN10 ) = 1-exp( Phred * -LOG10_E_tenth )
    for i in range( top1_bq_T.shape[0] ):
        lnL += log1p( -exp(-top1_bq_T[ i ]*LN10_tenth) )
    for i in range( top1_bq_C.shape[0] ):
        lnL += log1p( -exp(-top1_bq_C[ i ]*LN10_tenth) )
        
    for i in range( top2_bq_T.shape[0] ):
        lnL += log( exp(-top2_bq_T[ i ]*LN10_tenth) )
    for i in range( top2_bq_C.shape[0] ):
        lnL += log( exp(-top2_bq_C[ i ]*LN10_tenth) )        

    BIC = -2*lnL                # no free variable, no penalty
    return (lnL, BIC)

@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
cpdef tuple CalModel_Heter_noAS( np.ndarray[int32_t, ndim=1] top1_bq_T,np.ndarray[int32_t, ndim=1] top1_bq_C,np.ndarray[int32_t, ndim=1] top2_bq_T,np.ndarray[int32_t, ndim=1] top2_bq_C ):
    """Return (lnL, BIC)

    k_T
    k_C
    """
    cdef:
        int k_T, k_C
        double lnL, BIC
        int i
        int tn_T, tn_C, tn  # total observed NTs
        double lnL_T, lnL_C # log likelihood for treatment and control

    lnL = 0
    BIC = 0
    #for k_T
    # total oberseved treatment reads from top1 and top2 NTs
    tn_T = top1_bq_T.shape[0] + top2_bq_T.shape[0]
    
    if tn_T == 0:
        raise Exception("Total number of treatment reads is 0!")
    else:
        ( lnL_T, k_T ) = GreedyMaxFunctionNoAS( top1_bq_T.shape[0], top2_bq_T.shape[0], tn_T, top1_bq_T, top2_bq_T )
        lnL += lnL_T
        BIC += -2*lnL_T

    #for k_C
    tn_C = top1_bq_C.shape[0] + top2_bq_C.shape[0]

    if tn_C == 0:
        pass
    else:
        ( lnL_C, k_C )  = GreedyMaxFunctionNoAS( top1_bq_C.shape[0], top2_bq_C.shape[0], tn_C, top1_bq_C, top2_bq_C )
        lnL += lnL_C
        BIC += -2*lnL_C

    tn = tn_C + tn_T

    # we penalize big model depending on the number of reads/samples
    if tn_T == 0:
        BIC += log( tn_C )
    elif tn_C == 0:
        BIC += log( tn_T )
    else:
        BIC += log( tn_T ) + log( tn_C )

    return ( lnL, BIC )


@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
cpdef tuple CalModel_Heter_AS( np.ndarray[int32_t, ndim=1] top1_bq_T, np.ndarray[int32_t, ndim=1] top1_bq_C, np.ndarray[int32_t, ndim=1] top2_bq_T, np.ndarray[int32_t, ndim=1] top2_bq_C, float max_allowed_ar = 0.99 ):
    """Return (lnL, BIC)

    kc
    ki
    AS_alleleratio
    """
    cdef:
        int k_T, k_C
        double lnL, BIC
        int i
        int tn_T, tn_C, tn  # total observed NTs
        double lnL_T, lnL_C # log likelihood for treatment and control
        double AS_alleleratio   # allele ratio

    lnL = 0
    BIC = 0

    #assert top2_bq_T.shape[0] + top2_bq_C.shape[0] > 0, "Total number of top2 nt should not be zero while using this function: CalModel_Heter_AS!"

    # Treatment
    tn_T = top1_bq_T.shape[0] + top2_bq_T.shape[0]

    if tn_T == 0:
        raise Exception("Total number of treatment reads is 0!")
    else:
        ( lnL_T, k_T, AS_alleleratio ) = GreedyMaxFunctionAS( top1_bq_T.shape[0], top2_bq_T.shape[0], tn_T, top1_bq_T, top2_bq_T, max_allowed_ar)
        #print ">>>",lnL_T, k_T, AS_alleleratio
        lnL += lnL_T
        BIC += -2*lnL_T

    # control
    tn_C = top1_bq_C.shape[0] + top2_bq_C.shape[0]

    if tn_C == 0:
        pass
    else:
        # We assume control will not have allele preference
        ( lnL_C, k_C ) = GreedyMaxFunctionNoAS ( top1_bq_C.shape[0], top2_bq_C.shape[0], tn_C, top1_bq_C, top2_bq_C)
        lnL += lnL_C
        BIC += -2*lnL_C

    tn = tn_C + tn_T

    # we penalize big model depending on the number of reads/samples
    if tn_T == 0:
        BIC += log( tn_C )
    elif tn_C == 0:
        BIC += 2 * log( tn_T )
    else:
        BIC += 2 * log( tn_T ) + log( tn_C )
    
    return (lnL, BIC)


@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
cdef tuple GreedyMaxFunctionAS( int m, int n, int tn, np.ndarray[int32_t, ndim=1] me, np.ndarray[int32_t, ndim=1] ne, float max_allowed_ar = 0.99 ):
    """Return lnL, k and alleleratio in tuple.

    Note: I only translate Liqing's C++ code into pyx here. Haven't done any review.
    """
    cdef:
        double dnew, dold, rold, rnew
        int kold, knew
        bool btemp
        int k0
        double dl, dr, d0, d1l, d1r

    assert m+n == tn
    btemp = False
    if tn == 1:                 # only 1 read; I don't expect this to be run...
        dl=calculate_ln(m,n,tn,me,ne,0,0);
        dr=calculate_ln(m,n,tn,me,ne,1,1);

        if dl>dr:
            k = 0
            return ( dl, 0, 0 )
        else:
            k = 1
            return ( dr, 1, 1 )
    elif m == 0:                          #no top1 nt
        return ( calculate_ln( m, n, tn, me, ne, 0, m, max_allowed_ar ), m, 1-max_allowed_ar )
        #k0 = m + 1
    elif m == tn:                         #all reads are top1
        return ( calculate_ln( m, n, tn, me, ne, 1, m, max_allowed_ar ), m, max_allowed_ar )
    else:
        k0 = m

    d0  = calculate_ln( m, n, tn, me, ne, float(k0)/tn, k0, max_allowed_ar )
    d1l = calculate_ln( m, n, tn, me, ne, float(k0-1)/tn, k0-1, max_allowed_ar )
    d1r = calculate_ln( m, n, tn, me, ne, float(k0+1)/tn, k0+1, max_allowed_ar )

    if d0 > d1l-1e-8 and d0 > d1r-1e-8:
        k = k0
        ar = float(k0)/tn
        return ( d0, k, ar )
    elif d1l > d0:
        dold = d1l
        kold = k0-1
        rold = float(k0-1)/tn
        while kold > 1: #disable: when kold=1 still run, than knew=0 is the final run
            knew = kold - 1
            rnew = float(knew)/tn

            dnew = calculate_ln( m,n,tn,me,ne,rnew,knew, max_allowed_ar )

            if(dnew-1e-8 < dold) :
                btemp=True
                break
            kold=knew
            dold=dnew
            rold=rnew

        if btemp: #maximum L value is in [1,m-1];
            k = kold 
            ar= rold
            return ( dold, k, ar )
        else: #L(k=0) is the max for [0,m-1]
            k = kold
            ar = rold
            return ( dold, k, ar )

    elif d1r > d0:
        dold = d1r
        kold = k0 + 1
        rold = float(k0 + 1)/tn
        while kold < tn - 1: #//disable: when kold=tn-1 still run, than knew=tn is the final run
            knew = kold + 1
            
            rnew = float(knew)/tn
            
            dnew = calculate_ln( m,n,tn,me,ne,rnew,knew, max_allowed_ar )
            
            if dnew - 1e-8 < dold: 
                btemp = True
                break
            kold = knew
            dold = dnew
            rold = rnew

        if btemp: #maximum L value is in [m+1,tn-1]
            k = kold
            ar= rold
            return ( dold, k, ar )
        else: #L(k=tn) is the max for [m+1,tn]
            k = kold
            ar = rold
            return ( dold, k, ar )
    else:
        raise Exception("error in GreedyMaxFunctionAS")


@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
cdef tuple GreedyMaxFunctionNoAS (int m, int n, int tn, np.ndarray[int32_t, ndim=1] me, np.ndarray[int32_t, ndim=1] ne ):
    """Return lnL, and k in tuple.

    Note: I only translate Liqing's C++ code into pyx here. Haven't done any review.
    """
    cdef:
        double dnew, dold
        int kold, knew
        bool btemp
        int k0
        double bg_r, dl, dr, d0, d1l, d1r

    btemp = False
    bg_r = 0.5

    if tn == 1:
        dl = calculate_ln( m, n, tn, me, ne, bg_r, 0)
        dr= calculate_ln( m, n, tn, me, ne, bg_r, 1)
        if dl > dr:
            k = 0
            return ( dl, 0 )
        else:
            k = 1
            return ( dr, 1 )
    elif m == 0:                          #no top1 nt
        return ( calculate_ln( m, n, tn, me, ne, bg_r, m ), m )
        #k0 = m + 1
    elif m == tn:                         #all reads are top1
        return ( calculate_ln( m, n, tn, me, ne, bg_r, m ), m )
    #elif m == 0:
    #    k0 = m + 1
    #elif m == tn:
    #    k0 = m - 1
    else:
        k0 = m

    d0  = calculate_ln( m, n, tn, me, ne, bg_r, k0)
    d1l = calculate_ln( m, n, tn, me, ne, bg_r, k0 - 1)
    d1r = calculate_ln( m, n, tn, me, ne, bg_r, k0 + 1)

    if d0 > d1l - 1e-8 and d0 > d1r - 1e-8:
        k = k0
        return ( d0, k )
    elif d1l > d0:
        dold = d1l
        kold=k0 - 1
        while kold >= 1:   #//when kold=1 still run, than knew=0 is the final run
            knew = kold - 1
            dnew = calculate_ln( m, n, tn, me, ne, bg_r, knew )
            if dnew - 1e-8 < dold:
                btemp = True
                break
            kold=knew
            dold=dnew
                        
        if btemp: #//maximum L value is in [1,m-1];
            k = kold
            return ( dold, k )
        else: #//L(k=0) is the max for [0,m-1]
            k = kold
            return ( dold, k )
    elif d1r > d0:
        dold = d1r
        kold = k0 + 1
        while kold <= tn - 1: #//when kold=tn-1 still run, than knew=tn is the final run
            knew = kold + 1
            dnew = calculate_ln( m, n, tn, me, ne, bg_r, knew )
            if dnew - 1e-8 < dold:
                btemp = True
                break
            kold = knew
            dold = dnew

        if btemp: #//maximum L value is in [m+1,tn-1]
            k = kold
            return ( dold, k )
        else: #//L(k=tn) is the max for [m+1,tn]
            k = kold
            return ( dold, k )
    else:
        raise Exception("error in GreedyMaxFunctionNoAS")

@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
cdef calculate_ln( int m, int n, int tn, np.ndarray[int32_t, ndim=1] me, np.ndarray[int32_t, ndim=1] ne, double r, int k, float max_allowed_r = 0.99):
    """Calculate log likelihood given quality of top1 and top2, the ratio r and the observed k.

    """
    cdef:
        int i
        double lnL
        double e

    lnL = 0

    if r > max_allowed_r or r < 1 - max_allowed_r: # r is extremely high or
        #print "here1"
        lnL += k*log( max_allowed_r ) + (tn-k)*log( 1- max_allowed_r) #-10000
    else:
        lnL += k*log( r ) + (tn-k)*log(1-r)

    if k == 0 or k == tn:       # it's entirely biased toward 1 allele
        #print "here2"
        pass
        #lnL += k*log( max_allowed_r ) #-10000
        #lnL += -10000
    elif k <= tn/2:
        for i in range( k ):
            lnL += log(float(tn-i)/(k-i))
    else:
        for i in range( tn-k ):
            lnL += log(float(tn-i)/(tn-k-i))

    for i in range( m ):
        e = exp( - me[ i ] * LN10_tenth )
        lnL += log((1-e)*(float(k)/tn) + e*(1-float(k)/tn))
    
    for i in range( n ):
        e = exp( - ne[ i ] * LN10_tenth )
        lnL += log((1-e)*(1-float(k)/tn) + e*(float(k)/tn))

    #print r,k,lnL
    return lnL

cpdef int calculate_GQ ( double lnL1, double lnL2, double lnL3):
    """GQ1 = -10*log_{10}((L2+L3)/(L1+L2+L3))
       
       
    """
    cdef:
        double L1, L2, L3, sum, tmp
        int GQ_score
        
    #L1 = exp(lnL1-lnL1)
    L1 = 1
    L2 = exp(lnL2-lnL1)
    L3 = exp(lnL3-lnL1)

    #if L1 > 1:
    #    L1 = 1

    if L2 > 1:
        L2 = 1
    if L3 > 1:
        L3 = 1
    #if(L1<1e-110) L1=1e-110;
    if L2 < 1e-110:
        L2=1e-110
    if L3 < 1e-110:
        L3 = 1e-110

    sum = L1 + L2 + L3
    tmp = ( L2 + L3 )/sum
    if tmp > 1e-110:
        GQ_score = (int)(-4.34294*log(tmp))
    else:
        GQ_score = 255

    return GQ_score;

cpdef int calculate_GQ_heterASsig( double lnL1, double lnL2):
    """
    """
    cdef:
        double L1, L2, sum, tmp
        int ASsig_score

    #L1=exp(2.7182818,lnL1-lnL1)
    L1 = 1
    L2 = exp( lnL2 - lnL1 )

    #if L1 > 1:
    #    L1 = 1
    if L2 > 1:
        L2 = 1
    #if L1 < 1e-110:
    #    L1 = 1e-110
    if L2 < 1e-110:
        L2 = 1e-110

    sum = L1 + L2
    tmp = L2/sum
    if tmp > 1e-110:
        ASsig_score = (int)(-4.34294*log(tmp))
    else:
        ASsig_score = 255

    return ASsig_score




MACS_VERSION = "3.0.0"
MAX_PAIRNUM = 1000
MAX_LAMBDA  = 100000
FESTEP      = 20
BUFFER_SIZE = 100000                   # np array will increase at step of 1 million items
READ_BUFFER_SIZE = 10000000            # 10M bytes for read buffer size
N_MP = 2                               # Number of processers

#Effective genome size, collected from
#https://deeptools.readthedocs.io/en/develop/content/feature/effectiveGenomeSize.html

EFFECTIVEGS = {"hs":2913022398,                  #GRCh38
               "mm":2652783500,                  #GRCm38
               "ce":100286401,                   #WBcel235
               "dm":142573017}                   #dm6



# Time-stamp: <2020-11-26 17:04:26 Tao Liu>

"""Description: Naive call differential peaks from 4 bedGraph tracks for scores.

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

# ------------------------------------
# python modules
# ------------------------------------

import sys
import os
from MACS3.IO import BedGraphIO
from MACS3.Signal import ScoreTrack

# ------------------------------------
# constants
# ------------------------------------

# ------------------------------------
# Misc functions
# ------------------------------------
import logging
import MACS3.Utilities.Logger

logger = logging.getLogger(__name__)
debug   = logger.debug
info    = logger.info
error   = logger.critical
warn    = logger.warning
# ------------------------------------
# Classes
# ------------------------------------

# ------------------------------------
# Main function
# ------------------------------------
def run( options ):
    if options.maxgap >= options.minlen:
        error("MAXGAP should be smaller than MINLEN! Your input is MAXGAP = %d and MINLEN = %d" % (options.maxgap, options.minlen))

    LLR_cutoff = options.cutoff
    ofile_prefix = options.oprefix

    info("Read and build treatment 1 bedGraph...")
    t1bio = BedGraphIO.bedGraphIO(options.t1bdg)
    t1btrack = t1bio.build_bdgtrack()

    info("Read and build control 1 bedGraph...")
    c1bio = BedGraphIO.bedGraphIO(options.c1bdg)
    c1btrack = c1bio.build_bdgtrack()

    info("Read and build treatment 2 bedGraph...")
    t2bio = BedGraphIO.bedGraphIO(options.t2bdg)
    t2btrack = t2bio.build_bdgtrack()

    info("Read and build control 2 bedGraph...")
    c2bio = BedGraphIO.bedGraphIO(options.c2bdg)
    c2btrack = c2bio.build_bdgtrack()

    depth1 = options.depth1
    depth2 = options.depth2

    if depth1 > depth2:         # scale down condition 1 to size of condition 2
        depth1 = depth2 / depth1
        depth2 = 1.0
    elif depth1 < depth2:       # scale down condition 2 to size of condition 1
        depth2 = depth1/ depth2
        depth1 = 1.0
    else:                       # no need to scale down any
        depth1 = 1.0
        depth2 = 1.0

    twoconditionscore = ScoreTrack.TwoConditionScores( t1btrack,
                                                        c1btrack,
                                                        t2btrack,
                                                        c2btrack,
                                                        depth1,
                                                        depth2 )
    twoconditionscore.build()
    twoconditionscore.finalize()
    (cat1,cat2,cat3) = twoconditionscore.call_peaks(min_length=options.minlen, max_gap=options.maxgap, cutoff=options.cutoff)

    info("Write peaks...")

    if options.ofile:
        ofiles = [os.path.join( options.outdir, x ) for x in options.ofile]
        name_prefix = [ x.encode() for x in options.ofile ]
    else:
        ofiles = [ os.path.join( options.outdir, "%s_c%.1f_cond1.bed" % (options.oprefix,options.cutoff)),
                   os.path.join( options.outdir, "%s_c%.1f_cond2.bed" % (options.oprefix,options.cutoff)),
                   os.path.join( options.outdir, "%s_c%.1f_common.bed" % (options.oprefix,options.cutoff))
                   ]
        name_prefix = [ x.encode() for x in [ options.oprefix+"_cond1_", options.oprefix+"_cond2_", options.oprefix+"_common_" ]]

    nf = open( ofiles[ 0 ], 'w' )
    cat1.write_to_bed(nf, name_prefix=name_prefix[ 0 ], name=b"condition 1", description=b"unique regions in condition 1", score_column="score")
    nf.close()

    nf = open( ofiles[ 1 ], 'w' )
    cat2.write_to_bed(nf, name_prefix=name_prefix[ 1 ], name=b"condition 2", description=b"unique regions in condition 2", score_column="score")
    nf.close()

    nf = open( ofiles[ 2 ], 'w' )
    cat3.write_to_bed(nf, name_prefix=name_prefix[ 2 ], name=b"common", description=b"common regions in both conditions", score_column="score")
    nf.close()
    info("Done")




# Time-stamp: <2020-11-24 17:00:16 Tao Liu>

"""Description: Random sample certain number/percentage of tags.

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

# ------------------------------------
# python modules
# ------------------------------------

import os
import sys

# ------------------------------------
# own python modules
# ------------------------------------
from MACS3.Utilities.Constants import *
from MACS3.Utilities.OptValidator import opt_validate_randsample

# ------------------------------------
# Main function
# ------------------------------------
def run( options0 ):
    options = opt_validate_randsample( options0 )
    # end of parsing commandline options
    info = options.info
    warn = options.warn
    debug = options.debug
    error = options.error

    options.PE_MODE = options.format in ('BAMPE','BEDPE')

    #0 check output file
    if options.outputfile:
        outfhd = open( os.path.join( options.outdir, options.outputfile ), "w" )
    else:
        outfhd = sys.stdout

    #1 Read tag files
    if options.PE_MODE:
        info("# read input file in Paired-end mode.")
        treat = load_frag_files_options ( options ) # return PETrackI object
        t0 = treat.total # total fragments
        info("# total fragments/pairs in alignment file: %d" % (t0) )
    else:
        info("read tag files...")
        treat = load_tag_files_options (options)

        info("tag size = %d" % options.tsize)
        treat.fw = options.tsize

        t0 = treat.total
        info(" total tags in alignment file: %d" % (t0))

    if options.number:
        if options.number > t0:
            error(" Number you want is bigger than total number of tags in alignment file! Please specify a smaller number and try again!")
            error(" %.2e > %.2e" % (options.number, t0))
            sys.exit(1)
        info(" Number of tags you want to keep: %.2e" % (options.number))
        options.percentage = float(options.number)/t0*100
    info(" Percentage of tags you want to keep: %.2f%%" % (options.percentage))

    if options.seed >= 0:
        info(" Random seed has been set as: %d" % options.seed )

    treat.sample_percent(options.percentage/100.0, options.seed )

    info(" tags after random sampling in alignment file: %d" % (treat.total))

    info("Write to BED file")
    treat.print_to_bed(fhd=outfhd)
    info("finished! Check %s." % options.outputfile)

def load_tag_files_options ( options ):
    """From the options, load alignment tags.

    """
    options.info("# read treatment tags...")
    tp = options.parser(options.ifile[0], buffer_size=options.buffer_size)
    if not options.tsize:           # override tsize if user specified --tsize
        ttsize = tp.tsize()
        options.tsize = ttsize
    treat = tp.build_fwtrack()
    #treat.sort()
    if len(options.ifile) > 1:
        # multiple input
        for ifile in options.ifile[1:]:
            tp = options.parser(ifile, buffer_size=options.buffer_size)
            treat = tp.append_fwtrack( treat )
            #treat.sort()
    treat.finalize()

    options.info("tag size is determined as %d bps" % options.tsize)
    return treat

def load_frag_files_options ( options ):
    """From the options, load treatment fragments and control fragments (if available).

    """
    options.info("# read treatment fragments...")

    tp = options.parser(options.ifile[0], buffer_size=options.buffer_size)
    treat = tp.build_petrack()
    #treat.sort()
    if len(options.ifile) > 1:
        # multiple input
        for ifile in options.ifile[1:]:
            tp = options.parser(ifile, buffer_size=options.buffer_size)
            treat = tp.append_petrack( treat )
            #treat.sort()
    treat.finalize()
    return treat



# cython: language_level=3
# cython: profile=True
# Time-stamp: <2022-04-15 11:24:58 Tao Liu>

"""Module description:

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""
# ------------------------------------
# python modules
# ------------------------------------
from math import sqrt
import logging
import MACS3.Utilities.Logger

logger = logging.getLogger(__name__)
debug   = logger.debug
info    = logger.info
# ------------------------------------
# Other modules
# ------------------------------------

import numpy as np
cimport numpy as np
from cpython cimport bool

# ------------------------------------
# MACS3 modules
# ------------------------------------
#from scipy.stats import norm    # there is another implemented function MACS3.Signal.Prob.pnorm
from MACS3.Signal.Prob import pnorm2

# ------------------------------------
# Misc functions
# ------------------------------------
cdef tuple online_update( float x, long c, float m, float s):
    cdef:
        float delta
    c += 1
    if c == 1:
        return( 1, x, float(0.0) )
    delta = x - m
    m += delta / c
    s += delta * (x - m)
    return (c, m, s)

# https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.norm.html
cdef inline float get_weighted_density( int x, float m, float v, w ):
    """Description:
    
    parameters:
      1. x: the observed value
      2. m: the mean of gaussian
      3. v: the variance of the gaussian
      4. w: the weight
    return value:
    """
    return w * pnorm2( float(x), m, v )

cdef int return_greater( np.ndarray data ):
    """
    Return the index of the largest value in an array of doubles
    @param data: an Array of doubles 
    @return an integer representing thre index of the largest value in the inputted array
    """
    cdef:
        int largest_index
        np.ndarray largest_inds
    largest_inds = np.argwhere(data == np.amax(data)) # find the indices of the most likely category(s) (max values)
    if len(largest_inds) > 1:    # if there are multiple "most-likely" categories, ignore data point
        largest_index = -1
    else:
        largest_index = largest_inds[0][0]
    return largest_index
    
# ------------------------------------
# Classes
# ------------------------------------

cdef class HMMR_EM:
    """ Main HMMR EM class.

    This EM trainer will find the optimal mean and stddev of three of
    the four modes -- mono-nucleosomal, di-nucloeosomal, and
    three-nucleosomal fragments. Please note that the mean and stddev
    of the short fragment won't be optimized in this approach and only
    rely on the user's input.
    
    """
    cdef:
        public np.ndarray fragMeans    # fragment length mean for each of the three modes
        public np.ndarray fragStddevs  # fragment length standard deviation for each of the three modes
        np.ndarray fragVars            # we need variances to calculate PDF of normal distribution
        int min_fraglen
        int max_fraglen
        float epsilon          # maximum difference to call the value is converged
        int maxIter             # maximum iternation
        float jump
        int seed                # random seed for downsampling
        bool converged          # wheter the EM is converged
        float sample_percentage
        object __petrack          # PETrackI object
        np.ndarray __data         # data for fragment lengths
        np.ndarray __weights
        
    def __init__ ( self, object petrack, list init_means, list init_stddevs , int min_fraglen = 100, int max_fraglen = 1000, float sample_percentage  = 10, float epsilon = 0.05, int maxIter = 20, float jump = 1.5, int seed = 12345):
        """Initialize HMMR_EM object. The first three parameters are required.

        parameters:
            1. petrack: a MACS3.Signal.PairedEndTrack.PETrackI object
            2. init_means: list of initial means of fragments, for mono, di, and tri signals
            3. init_stddevs: list of initial stddevs of fragments, for mono, di, and tri signals
            4. min_fraglen
            5. max_fraglen
            6. sample_percentage: downsample the original data to get the lengths distribution, default 10
            7. epsilon
            8. maxIter
            9. jump: to emplify the difference
            9. seed
        """
        cdef:
            float cutoff1, cutoff2
            long sum1, sum2, sum3, counter1, counter2, counter3
        # initial values
        self.min_fraglen = min_fraglen
        self.max_fraglen = max_fraglen
        self.epsilon = epsilon
        self.maxIter = maxIter
        self.jump = jump
        self.seed = seed
        self.converged = False
        self.fragMeans = np.array(init_means, dtype=float)
        self.fragStddevs = np.array(init_stddevs, dtype=float)
        self.fragVars = (np.array(init_stddevs, dtype=float)) ** 2
        self.sample_percentage = sample_percentage

        # first, let's prepare the lengths data
        # sample down
        self.__petrack = petrack.sample_percent_copy( self.sample_percentage/100.0, seed = self.seed ) # may need to provide seed option for init function
        self.__data = self.__petrack.fraglengths()
        #fhd = open("allfrag.txt","w")
        #for d in self.__data:
        #    fhd.write(f"{d}\n")
        #fhd.close()
        # then we only keep those with fragment lengths within certain range
        self.__data = self.__data[ np.logical_and( self.__data >= self.min_fraglen, self.__data <= self.max_fraglen ) ]

        info(f"# Downsampled {self.__data.shape[0]} fragments will be used for EM training...")
        # next, we will calculate the weights -- ie the proportion of fragments in each length category
        cutoff1 = (init_means[ 1 ] - init_means[ 0 ])/2 + init_means[ 0 ]
        cutoff2 = (init_means[ 2 ] - init_means[ 1 ])/2 + init_means[ 1 ]

        sum3 = len( self.__data )
        sum2 = sum( self.__data < cutoff2 )
        sum1 = sum( self.__data < cutoff1 )

        counter3 = sum3 - sum2
        counter2 = sum2 - sum1
        counter1 = sum1

        self.__weights = np.array([ counter1/sum3, counter2/sum3, counter3/sum3])
        debug( f"initial: means: {self.fragMeans}, stddevs: {self.fragStddevs}, weights: {self.__weights}" )
        self.__learn()
        return

    cdef bool __learn(self):
        """Description: When we train the mean and stddev for 

        parameters:
        return value:
        """
        cdef:
            int itr = 0         # number of iterations
            int i
            int counter         # number of modes that has been converged
            np.ndarray old_means, old_stddevs, old_weights

        old_means = np.zeros( 3, dtype=float )
        old_stddevs = np.zeros( 3, dtype=float )
        old_weights = np.zeros( 3, dtype=float )
        
        self.converged = False
        while self.converged == False:
            for i in range( 3 ):                
                old_means[i] = self.fragMeans[i]
                old_stddevs[i] = self.fragStddevs[i]
                old_weights[i] = self.__weights[i]

            self.__iterate()
            itr += 1
            debug( f"after iteration {itr}: means: {self.fragMeans}, stddevs: {self.fragStddevs}, weights: {self.__weights}" )
            counter = 0
            for i in range( 3 ):
                if abs(old_means[i] - self.fragMeans[i]) < self.epsilon and abs(old_weights[i] - self.__weights[i]) < self.epsilon and abs(old_stddevs[i] - self.fragStddevs[i]) < self.epsilon:
                    counter += 1
            if counter == 3:
                self.converged = True
                info( f"# Reached convergence after {itr} iterations" )
            if itr >= self.maxIter:
                info( f"# Reached maximum number ({self.maxIter}) of iterations" )
                break
        return self.converged

    cdef void __iterate(self):
        """Description: This is a private function only used by HMMR_EM class

        parameters:
        return value:
        """
        cdef:
            np.ndarray temp, counter, means, variances, s, c
            long total
            int i, j, index

        temp = np.zeros(3, dtype=float) # for each category, the likelihood
        counter = np.zeros(3, dtype=int) # for each category, the number of data points/fragment
        total = 0 # total number of data points/fragments assigned to
                  # three categories
        means = np.zeros(3, dtype=float) # for each category, the new mean
        variances = np.zeros(3, dtype=float) # for each category, the new variances
        s = np.zeros(3, dtype=float)         # for each category, __s
                                             # is for storing
                                             # intermediate values for
                                             # the online algorithm
        c = np.zeros(3, dtype=long)          # for each category, __c
                                             # is for storing
                                             # intermediate values for
                                             # the online algorithm
        for i in range( len( self.__data ) ):
            for j in range( 3 ):
                # for each category: mono, di, tri- (3 in total), we get the likelihoods
                #debug( f"j:{j} {self.__data[i]}, {self.fragMeans[j]}, {self.fragVars[j]}, {self.__weights[j]}" )
                temp[j] = get_weighted_density( self.__data[i], self.fragMeans[j], self.fragVars[j], self.__weights[j] )
            # now look for the most likely category, as `index`
            index = return_greater( temp )

            # then we will update __means and __stds
            if index != -1: # If we can find a mostly likely category
                ##---- update with online algorithm --
                (c[ index ], means[ index ], s[ index ]) = online_update( self.__data[ i ], c[ index ], means[ index ], s[ index ] )
                variances[index] = s[ index ]/c[ index ]
                total += 1
            #debug( f"index: {index} mean: {means[index]}" )

        # Set fragMeans, fragStddevs, and fragVars
        #debug( f"counts: {c[0]} {c[1]} {c[2]}" )
        #debug( f"means: {means[0]} {means[1]} {means[2]}" )
        #debug( f"vars: {variances[0]} {variances[1]} {variances[2]}")
        try:
            for j in range( 3 ):
                if c[ j ] == 0:
                    # if there is no fragment size in this category, ignore
                    continue
                self.fragMeans[ j ] = self.fragMeans[ j ] + self.jump*(means[ j ] - self.fragMeans[ j ])
                self.fragVars[ j ] = self.fragVars[ j ] + self.jump*(variances[ j ] - self.fragVars[ j ])
                self.fragStddevs[ j ] = sqrt( self.fragVars[ j ] )
                self.__weights[ j ] = c[ j ] / total
        except ValueError:
            print(' ValueError:  Adjust --means and --stddev options and re-run command')
        debug( f"After this iteration, {total} fragments have been assigned with either of the three modes" )
        return



# cython: language_level=3
# cython: profile=True
# Time-stamp: <2020-12-03 20:00:30 Tao Liu>

"""Module Description: functions to find maxima minima or smooth the
signal tracks.

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""
# ------------------------------------
# Python modules
# ------------------------------------
from math import factorial as mathfactorial
from math import sqrt as mathsqrt

# ------------------------------------
# Other modules
# ------------------------------------
# smoothing function
import numpy as np
cimport numpy as np
from numpy cimport uint8_t, uint16_t, uint32_t, uint64_t, int8_t, int16_t, int32_t, int64_t, float32_t, float64_t
from cpython cimport bool


cpdef np.ndarray[int32_t, ndim=1] maxima(np.ndarray[float32_t, ndim=1] signal,
                                            int window_size=51):
    """return the local maxima in a signal after applying a 2nd order
    Savitsky-Golay (polynomial) filter using window_size specified
    """
    cdef:
        np.ndarray[int32_t, ndim=1] m
        np.ndarray[float64_t, ndim=1] smoothed
        np.ndarray[float64_t, ndim=1] sign, diff

    window_size = window_size//2*2+1 # to make an odd number
    smoothed = savitzky_golay_order2_deriv1(signal, window_size).round(16)
    sign = np.sign( smoothed )
    diff = np.diff( sign )
    m = np.where( diff <= -1)[0].astype("int32")
    return m

cdef np.ndarray[int32_t, ndim=1] internal_minima( np.ndarray[float32_t, ndim=1] signal,
                                                  np.ndarray[int32_t, ndim=1] maxima ):
    cdef:
        np.ndarray[int32_t, ndim=1] ret
        int32_t n = maxima.shape[0]
        int32_t i, v, v2
    if n == 0 or n == 1:
        ret = np.ndarray(0, 'int32')
        return ret
    else:
        ret = np.zeros(n - 1, 'int32')
        pos1 = maxima[0]
        for i in range(n - 1):
            pos2 = maxima[i + 1]
            ret[i] = np.where(signal[pos1:pos2] == signal[pos1:pos2].min())[0][0] + pos1
            pos1 = pos2
        return ret

cdef inline float32_t sqrt(float32_t threshold):
    return mathsqrt(threshold)

cpdef enforce_peakyness(np.ndarray[float32_t, ndim=1] signal,
                        np.ndarray[int32_t, ndim=1] maxima):
    """requires peaks described by a signal and a set of points where the signal
    is at a maximum to meet a certain set of criteria

    maxima which do not meet the required criteria are discarded

    criteria:
        for each peak:
            calculate a threshold of the maximum of its adjacent two minima
                plus the sqrt of that value
            subtract the threshold from the region bounded by those minima
            clip that region if negative values occur inside it
            require it be > 50 bp in width
            require that it not be too flat (< 6 unique values)
    """
    cdef:
        np.ndarray[int32_t, ndim=1] minima = internal_minima(signal, maxima)
        np.ndarray[float32_t, ndim=1] new_signal
        int32_t n = minima.shape[0]
        float32_t threshold
        np.ndarray[int32_t, ndim=1] peaky_maxima = maxima.copy()
        int32_t j = 0
    if n == 0: return maxima
#    else:
    threshold = signal[minima[0]]
    threshold += sqrt(threshold)
    new_signal = signal[0:minima[0]] - threshold - sqrt(threshold)
#    assert maxima[0] < minima[0], '%d > %d' % ( maxima[0], minima[0] )
    if is_valid_peak(new_signal, maxima[0]):
        peaky_maxima[0] = maxima[0]
        j += 1
    for i in range(n - 1):
        threshold = max(signal[minima[i]], signal[minima[i + 1]])
        threshold += sqrt(threshold)
        new_signal = signal[minima[i]:minima[i+1]] - threshold
        new_maximum = maxima[i+1] - minima[i]
        if is_valid_peak(new_signal, new_maximum):
            peaky_maxima[j] = maxima[i + 1]
            j += 1
    threshold =  signal[minima[-1]]
    threshold += sqrt(threshold)
    new_signal = signal[minima[-1]:] - threshold
    new_maximum = maxima[-1] - minima[-1]
    if is_valid_peak(new_signal, new_maximum):
        peaky_maxima[j] = maxima[-1]
        j += 1
    peaky_maxima.resize(j, refcheck=False)
    return peaky_maxima

# hardcoded minimum peak width = 50
cdef bool is_valid_peak(np.ndarray[float32_t, ndim=1] signal, int maximum):
    cdef:
        np.ndarray s
        int32_t length
    s = hard_clip(signal, maximum)
    length = s.shape[0]
    if length < 50:
        return False
    elif too_flat(s):
        return False
    return True

# require at least 6 different float values -- prevents broad flat peaks
cdef bool too_flat(np.ndarray[float32_t, ndim=1] signal):
    """return whether signal has at least 6 unique values
    """
    return np.unique(signal).shape[0] < 6

# hard clip a region with negative values
cdef np.ndarray[float32_t, ndim=1] hard_clip(np.ndarray[float32_t, ndim=1] signal, int32_t maximum):
    """clip the signal in both directions at the nearest values <= 0
    to position maximum
    """
    cdef:
        int32_t i
        int32_t left = 0
        int32_t right = signal.shape[0]
    # clip left
    for i in range( right - maximum, 0 ):
        if signal[ -i ] < 0:
            left = i
            break
    for i in range(maximum, right):
        if signal[i] < 0:
            right = i
            break
    return signal[ left:right ]

cpdef np.ndarray[ int32_t, ndim=1 ] enforce_valleys(np.ndarray[ float32_t, ndim=1 ] signal,
                                                    np.ndarray[ int32_t, ndim=1 ] summits,
                                                    float32_t min_valley = 0.8 ):
    """require a value of <= min_valley * lower summit
    between each pair of summits
    """
    cdef:
        float32_t req_min, v, prev_v
        int32_t summit_pos, prev_summit_pos
        int32_t n_summits
        int32_t n_valid_summits
        np.ndarray[ int32_t, ndim=1 ] valid_summits
    n_summits = summits.shape[0]
    n_valid_summits = 1    
    valid_summits = summits.copy()        
    # Remove peaks that do not have sufficient valleys
    if n_summits == 1: return summits
    for i in range( 1, n_summits ):
        prev_summit_pos = valid_summits[ n_valid_summits-1 ]
        summit_pos = summits[ i ]
        prev_v = signal[ prev_summit_pos ]
        v = signal[ summit_pos ]
        req_min = min_valley * min( prev_v, v )
        if ( signal[ prev_summit_pos:summit_pos ] < req_min ).any():
            valid_summits[ n_valid_summits ] = summit_pos
            n_valid_summits += 1
        elif v > prev_v:
            valid_summits[ n_valid_summits-1 ] = summit_pos
    valid_summits.resize( n_valid_summits, refcheck=False )
    return valid_summits

# Modified from http://www.scipy.org/Cookbook/SavitzkyGolay
# positive window_size not enforced anymore
# needs sane input paramters, window size > 4
# switched to double precision for internal accuracy
cpdef np.ndarray[float64_t, ndim=1] savitzky_golay_order2_deriv1(np.ndarray[float32_t, ndim=1] signal,
                                                                 int32_t window_size):
    """Smooth (and optionally differentiate) data with a Savitzky-Golay filter.
    The Savitzky-Golay filter removes high frequency noise from data.
    It has the advantage of preserving the original shape and
    features of the signal better than other types of filtering
    approaches, such as moving averages techhniques.
    Parameters
    ----------
    y : array_like, shape (N,)
        the values of the time history of the signal.
    window_size : int
        the length of the window. Must be an odd integer number.
    deriv: int
        the order of the derivative to compute (default = 0 means only smoothing)
    Returns
    -------
    ys : ndarray, shape (N)
        the smoothed signal (or it's n-th derivative).
    Notes
    -----
    The Savitzky-Golay is a type of low-pass filter, particularly
    suited for smoothing noisy data. The main idea behind this
    approach is to make for each point a least-square fit with a
    polynomial of high order over a odd-sized window centered at
    the point.

    References
    ----------
    .. [1] A. Savitzky, M. J. E. Golay, Smoothing and Differentiation of
       Data by Simplified Least Squares Procedures. Analytical
       Chemistry, 1964, 36 (8), pp 1627-1639.
    .. [2] Numerical Recipes 3rd Edition: The Art of Scientific Computing
       W.H. Press, S.A. Teukolsky, W.T. Vetterling, B.P. Flannery
       Cambridge University Press ISBN-13: 9780521880688
    """
    cdef:
        int32_t half_window, k
        np.ndarray[int64_t, ndim=2] b
        # pad the signal at the extremes with
        # values taken from the signal itself
        np.ndarray[float32_t, ndim=1] firstvals, lastvals
        np.ndarray[float64_t, ndim=1] m, ret

    if window_size % 2 != 1: window_size += 1
    half_window = (window_size - 1) // 2
    # precompute coefficients
    b = np.array([[1, k, k**2] for k in range(-half_window, half_window+1)],
                 dtype='int64')
    m = np.linalg.pinv(b)[1]
    # pad the signal at the extremes with
    # values taken from the signal itself
    firstvals = signal[0] - np.abs(signal[1:half_window+1][::-1] - signal[0])
    lastvals = signal[-1] + np.abs(signal[-half_window-1:-1][::-1] - signal[-1])
    signal = np.concatenate((firstvals, signal, lastvals))
    ret = np.convolve( m[::-1], signal.astype("float64"), mode='valid') #.astype("float32").round(8) # round to 8 decimals to avoid signing issue
    return ret

# Another modified version from http://www.scipy.org/Cookbook/SavitzkyGolay
cpdef np.ndarray[float32_t, ndim=1] savitzky_golay( np.ndarray[float32_t, ndim=1] y, int32_t window_size,
                                                    int32_t order, int32_t deriv = 0, int32_t rate = 1 ):
    """Smooth (and optionally differentiate) data with a Savitzky-Golay filter.
    The Savitzky-Golay filter removes high frequency noise from data.
    It has the advantage of preserving the original shape and
    features of the signal better than other types of filtering
    approaches, such as moving averages techniques.
    Parameters
    ----------
    y : array_like, shape (N,)
        the values of the time history of the signal.
    window_size : int
        the length of the window. Must be an odd integer number.
    order : int
        the order of the polynomial used in the filtering.
        Must be less then `window_size` - 1.
    deriv: int
        the order of the derivative to compute (default = 0 means only smoothing)
    Returns
    -------
    ys : ndarray, shape (N)
        the smoothed signal (or it's n-th derivative).
    Notes
    -----
    The Savitzky-Golay is a type of low-pass filter, particularly
    suited for smoothing noisy data. The main idea behind this
    approach is to make for each point a least-square fit with a
    polynomial of high order over a odd-sized window centered at
    the point.
    Examples
    --------
    t = np.linspace(-4, 4, 500)
    y = np.exp( -t**2 ) + np.random.normal(0, 0.05, t.shape)
    ysg = savitzky_golay(y, window_size=31, order=4)
    import matplotlib.pyplot as plt
    plt.plot(t, y, label='Noisy signal')
    plt.plot(t, np.exp(-t**2), 'k', lw=1.5, label='Original signal')
    plt.plot(t, ysg, 'r', label='Filtered signal')
    plt.legend()
    plt.show()
    References
    ----------
    .. [1] A. Savitzky, M. J. E. Golay, Smoothing and Differentiation of
       Data by Simplified Least Squares Procedures. Analytical
       Chemistry, 1964, 36 (8), pp 1627-1639.
    .. [2] Numerical Recipes 3rd Edition: The Art of Scientific Computing
       W.H. Press, S.A. Teukolsky, W.T. Vetterling, B.P. Flannery
       Cambridge University Press ISBN-13: 9780521880688
    """
    cdef:
        int32_t half_window, k
        np.ndarray[int64_t, ndim=2] b
        # pad the signal at the extremes with
        # values taken from the signal itself
        np.ndarray[float32_t, ndim=1] firstvals, lastvals, ret
        np.ndarray[float64_t, ndim=1] m

    try:
        window_size = np.abs( np.int( window_size ) )
        order = np.abs( np.int( order ) )
    except ValueError, msg:
        raise ValueError("window_size and order have to be of type int")
    if window_size % 2 != 1 or window_size < 1:
        raise TypeError("window_size size must be a positive odd number")
    if window_size < order + 2:
        raise TypeError("window_size is too small for the polynomials order")
    half_window = ( window_size -1 ) // 2
    # precompute coefficients
    b = np.array( [ [ k**i for i in range( order + 1 ) ] for k in range( -half_window, half_window+1 ) ] )
    m = np.linalg.pinv( b )[ deriv ] * rate**deriv * mathfactorial( deriv )
    # pad the signal at the extremes with
    # values taken from the signal itself
    firstvals = y[ 0 ] - np.abs( y[ 1:half_window + 1 ][ ::-1 ] - y[ 0 ] )
    lastvals = y[ -1 ] + np.abs( y[ -half_window - 1:-1 ][ ::-1 ] - y[ -1 ])
    y = np.concatenate( ( firstvals, y, lastvals ) )
    ret = np.convolve( m[ ::-1 ], y, mode = 'valid' ).astype("float32")
    return ret



# cython: language_level=3
# cython: profile=True
# Time-stamp: <2023-08-02 14:19:28 Tao Liu>

"""Module Description: For pileup functions.

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

# ------------------------------------
# python modules
# ------------------------------------
from time import time as ttime

# ------------------------------------
# MACS3 modules
# ------------------------------------
from MACS3.Utilities.Constants import *
from MACS3.Signal.cPosValCalculation cimport single_end_pileup as c_single_end_pileup
from MACS3.Signal.cPosValCalculation cimport write_pv_array_to_bedGraph as c_write_pv_array_to_bedGraph
from MACS3.Signal.cPosValCalculation cimport PosVal
from MACS3.Signal.cPosValCalculation cimport quick_pileup as c_quick_pileup

# ------------------------------------
# Other modules
# ------------------------------------
import numpy as np
cimport numpy as np
from numpy cimport int32_t, float32_t
from cpython cimport bool
from cpython cimport PyObject

# ------------------------------------
# C lib
# ------------------------------------
from libc.stdlib cimport malloc, free, qsort

# ------------------------------------
# utility internal functions
# ------------------------------------
cdef inline float mean( float a, float b ):
    return ( a + b ) / 2

cdef void clean_up_ndarray ( np.ndarray x ):
    """ Clean up numpy array in two steps
    """
    cdef:
        long i
    i = x.shape[0] // 2
    x.resize( 100000 if i > 100000 else i, refcheck=False)
    x.resize( 0, refcheck=False)
    return

cdef np.ndarray[np.int32_t, ndim=1] fix_coordinates(np.ndarray[np.int32_t, ndim=1] poss, int rlength):
    """Fix the coordinates.
    """
    cdef:
        long i
        int32_t * ptr

    ptr = <int32_t *> poss.data

    # fix those negative coordinates
    for i in range( poss.shape[0] ):
        if ptr[i] < 0:
            ptr[i] = 0
        else:
            break

    # fix those over-boundary coordinates
    for i in range( poss.shape[0]-1, -1, -1 ):
        if ptr[i] > rlength:
            ptr[i] = rlength
        else:
            break
    return poss

# ------------------------------------
# functions
# ------------------------------------

# ------------------------------------
# functions for pileup_cmd only
# ------------------------------------

# This function uses pure C code for pileup
cpdef void pileup_and_write_se( trackI,
                        bytes output_filename,
                        int d,
                        float scale_factor,
                        float baseline_value = 0.0,
                        bint directional = True,
                        bint halfextension = True ):
    """ Pileup a FWTrackI object and write the pileup result into a
    bedGraph file.

    This function calls pure C functions from cPosValCalculation.

    This function is currently only used `macs3 pileup` cmd.
    """
    cdef:
        long five_shift, three_shift, l, i
        list chroms
        int n_chroms
        bytes chrom
        np.ndarray[np.int32_t, ndim=1] plus_tags, minus_tags
        dict chrlengths = trackI.get_rlengths ()
        int * plus_tags_pos
        int * minus_tags_pos
        long rlength
        long fl
        bytes py_bytes
        char * chrom_char
        PosVal * _data
        long l_data

    # This block should be reused to determine the actual shift values
    if directional:
        # only extend to 3' side
        if halfextension:
            five_shift = d//-4  # five shift is used to move cursor towards 5' direction to find the start of fragment
            three_shift = d*3//4 # three shift is used to move cursor towards 3' direction to find the end of fragment
        else:
            five_shift = 0
            three_shift = d
    else:
        # both sides
        if halfextension:
            five_shift = d//4
            three_shift = five_shift
        else:
            five_shift = d//2
            three_shift = d - five_shift
    # end of the block

    chroms = list(chrlengths.keys())
    n_chroms = len( chroms )

    fh = open(output_filename, "w")
    fh.write("")
    fh.close()

    for i in range( n_chroms ):
        chrom = chroms[ i ]
        (plus_tags, minus_tags) = trackI.get_locations_by_chr(chrom)
        rlength = <long> chrlengths[ chrom ]
        plus_tags_pos = <int *> plus_tags.data
        minus_tags_pos = <int *> minus_tags.data

        _data = c_single_end_pileup( plus_tags_pos, plus_tags.shape[0], minus_tags_pos, minus_tags.shape[0], five_shift, three_shift, 0, rlength, scale_factor, baseline_value, &l_data )

        # write
        py_bytes = chrom
        chrom_char = py_bytes
        c_write_pv_array_to_bedGraph( _data, l_data, chrom_char, output_filename, 1 )

        # clean
        free( _data )
    return

# function to pileup BAMPE/BEDPE stored in PETrackI object and write to a BEDGraph file
# this function uses c function
cpdef pileup_and_write_pe( petrackI,
                           bytes output_filename,
                           float scale_factor = 1,
                           float baseline_value = 0.0):
    """ Pileup a PETrackI object and write the pileup result into a
    bedGraph file.

    This function calls pure C functions from cPosValCalculation.

    This function is currently only used `macs3 pileup` cmd.
    """
    cdef:
        dict chrlengths = petrackI.get_rlengths ()
        list chroms
        int n_chroms
        int i
        bytes chrom

        np.ndarray locs
        np.ndarray[np.int32_t, ndim=1] locs0
        np.ndarray[np.int32_t, ndim=1] locs1

        int * start_pos
        int * end_pos
        long fl
        bytes py_bytes
        char * chrom_char
        PosVal * _data
        long l_data

    chroms = list(chrlengths.keys())
    n_chroms = len( chroms )

    fh = open(output_filename, "w")
    fh.write("")
    fh.close()

    for i in range( n_chroms ):
        chrom = chroms[ i ]
        locs = petrackI.get_locations_by_chr(chrom)

        locs0 = np.sort(locs['l'])
        locs1 = np.sort(locs['r'])
        start_pos = <int *> locs0.data
        end_pos   = <int *> locs1.data

        _data = c_quick_pileup ( start_pos, end_pos, locs0.shape[0], scale_factor, baseline_value, &l_data )

        # write
        py_bytes = chrom
        chrom_char = py_bytes
        c_write_pv_array_to_bedGraph( _data, l_data, chrom_char, output_filename, 1 )

        # clean
        free( _data )        
    return

# ------------------------------------
# functions for other codes
# ------------------------------------

# general pileup function implemented in cython
cpdef list se_all_in_one_pileup ( np.ndarray[np.int32_t, ndim=1] plus_tags, np.ndarray[np.int32_t, ndim=1] minus_tags, long five_shift, long three_shift, int rlength, float scale_factor, float baseline_value ):
    """Return pileup given 5' end of fragment at plus or minus strand
    separately, and given shift at both direction to recover a
    fragment. This function is for single end sequencing library
    only. Please directly use 'quick_pileup' function for Pair-end
    library.

    It contains a super-fast and simple algorithm proposed by Jie
    Wang. It will take sorted start positions and end positions, then
    compute pileup values.

    It will return a pileup result in similar structure as
    bedGraph. There are two python arrays:

    [end positions, values] or '[p,v] array' in other description for
    functions within MACS3.

    Two arrays have the same length and can be matched by index. End
    position at index x (p[x]) record continuous value of v[x] from
    p[x-1] to p[x].

    """
    cdef:
        long i_s, i_e, i, I
        int a, b, p, pre_p, pileup
        list tmp
        long lp = plus_tags.shape[0]
        long lm = minus_tags.shape[0]
        long l = lp + lm
        np.ndarray[np.int32_t, ndim=1] start_poss, end_poss, ret_p
        np.ndarray[np.float32_t, ndim=1] ret_v
        # pointers are used for numpy arrays
        int32_t * start_poss_ptr
        int32_t * end_poss_ptr
        int32_t * ret_p_ptr     # pointer for position array
        float32_t * ret_v_ptr     # pointer for value array


    start_poss = np.concatenate( ( plus_tags-five_shift, minus_tags-three_shift ) )
    end_poss   = np.concatenate( ( plus_tags+three_shift, minus_tags+five_shift ) )

    # sort
    start_poss.sort()
    end_poss.sort()

    # fix negative coordinations and those extends over end of chromosomes
    start_poss = fix_coordinates(start_poss, rlength)
    end_poss = fix_coordinates(end_poss, rlength)

    lx = start_poss.shape[0]

    start_poss_ptr = <int32_t *> start_poss.data
    end_poss_ptr = <int32_t *> end_poss.data

    ret_p = np.zeros( 2 * lx, dtype="int32" )
    ret_v = np.zeros( 2 * lx, dtype="float32" )

    ret_p_ptr = <int32_t *> ret_p.data
    ret_v_ptr = <float32_t *> ret_v.data

    tmp = [ret_p, ret_v]        # for (endpos,value)
    i_s = 0                     # index of start_poss
    i_e = 0                     # index of end_poss
    I = 0

    pileup = 0
    if start_poss.shape[0] == 0: return tmp
    pre_p = min(start_poss_ptr[0],end_poss_ptr[0])

    if pre_p != 0:
        # the first chunk of 0
        ret_p_ptr[0] = pre_p
        ret_v_ptr[0] = max(0,baseline_value)
        ret_p_ptr += 1
        ret_v_ptr += 1
        I += 1

    pre_v = pileup

    assert start_poss.shape[0] == end_poss.shape[0]
    lx = start_poss.shape[0]

    while i_s < lx and i_e < lx:
        if start_poss_ptr[0] < end_poss_ptr[0]:
            p = start_poss_ptr[0]
            if p != pre_p:
                ret_p_ptr[0] = p
                ret_v_ptr[0] = max(pileup * scale_factor, baseline_value)
                ret_p_ptr += 1
                ret_v_ptr += 1
                I += 1
                pre_p = p
            pileup += 1
            i_s += 1
            start_poss_ptr += 1
        elif start_poss_ptr[0] > end_poss_ptr[0]:
            p = end_poss_ptr[0]
            if p != pre_p:
                ret_p_ptr[0] = p
                ret_v_ptr[0] = max(pileup * scale_factor, baseline_value)
                ret_p_ptr += 1
                ret_v_ptr += 1
                I += 1
                pre_p = p
            pileup -= 1
            i_e += 1
            end_poss_ptr += 1
        else:
            i_s += 1
            i_e += 1
            start_poss_ptr += 1
            end_poss_ptr += 1

    if i_e < lx:
        # add rest of end positions
        for i in range(i_e, lx):
            p = end_poss_ptr[0]
            if p != pre_p:
                ret_p_ptr[0] = p
                ret_v_ptr[0] = max(pileup * scale_factor, baseline_value)
                ret_p_ptr += 1
                ret_v_ptr += 1
                I += 1
                pre_p = p
            pileup -= 1
            end_poss_ptr += 1

    # clean mem
    clean_up_ndarray( start_poss )
    clean_up_ndarray( end_poss )

    # resize
    ret_p.resize( I, refcheck=False )
    ret_v.resize( I, refcheck=False )

    return tmp

# quick pileup implemented in cython
cpdef list quick_pileup ( np.ndarray[np.int32_t, ndim=1] start_poss, np.ndarray[np.int32_t, ndim=1] end_poss, float scale_factor, float baseline_value ):
    """Return pileup given plus strand and minus strand positions of fragments.

    A super-fast and simple algorithm proposed by Jie Wang. It will
    take sorted start positions and end positions, then compute pileup
    values.

    It will return a pileup result in similar structure as
    bedGraph. There are two python arrays:

    [end positions, values] or [p,v]

    Two arrays have the same length and can be matched by index. End
    position at index x (p[x]) record continuous value of v[x] from
    p[x-1] to p[x].

    """
    cdef:
        long i_s, i_e, i, I
        int a, b, p, pre_p, pileup
        list tmp
        long ls = start_poss.shape[0]
        long le = end_poss.shape[0]
        long l = ls + le
        np.ndarray[np.int32_t, ndim=1] ret_p
        np.ndarray[np.float32_t, ndim=1] ret_v
        # pointers are used for numpy arrays
        int32_t * start_poss_ptr
        int32_t * end_poss_ptr
        int32_t * ret_p_ptr     # pointer for position array
        float32_t * ret_v_ptr     # pointer for value array
        #int max_pileup = 0

    start_poss_ptr = <int32_t *> start_poss.data
    end_poss_ptr = <int32_t *> end_poss.data

    ret_p = np.zeros( l, dtype="int32" )
    ret_v = np.zeros( l, dtype="float32" )

    ret_p_ptr = <int32_t *> ret_p.data
    ret_v_ptr = <float32_t *> ret_v.data

    tmp = [ret_p, ret_v] # for (endpos,value)

    i_s = 0                         # index of plus_tags
    i_e = 0                         # index of minus_tags
    I = 0

    pileup = 0
    if ls == 0: return tmp
    pre_p = min(start_poss_ptr[0], end_poss_ptr[0])

    if pre_p != 0:
        # the first chunk of 0
        ret_p_ptr[0] = pre_p
        ret_v_ptr[0] = max(0,baseline_value)
        ret_p_ptr += 1
        ret_v_ptr += 1
        I += 1

    pre_v = pileup

    while i_s < ls and i_e < le:
        if start_poss_ptr[0] < end_poss_ptr[0]:
            p = start_poss_ptr[0]
            if p != pre_p:
                ret_p_ptr[0] = p
                ret_v_ptr[0] = max(pileup * scale_factor, baseline_value)
                ret_p_ptr += 1
                ret_v_ptr += 1
                I += 1
                pre_p = p
            pileup += 1
            #if pileup > max_pileup:
            #    max_pileup = pileup
            i_s += 1
            start_poss_ptr += 1
        elif start_poss_ptr[0] > end_poss_ptr[0]:
            p = end_poss_ptr[0]
            if p != pre_p:
                ret_p_ptr[0] = p
                ret_v_ptr[0] = max(pileup * scale_factor, baseline_value)
                ret_p_ptr += 1
                ret_v_ptr += 1
                I += 1
                pre_p = p
            pileup -= 1
            i_e += 1
            end_poss_ptr += 1
        else:
            i_s += 1
            i_e += 1
            start_poss_ptr += 1
            end_poss_ptr += 1

    if i_e < le:
        # add rest of end positions
        for i in range(i_e, le):
            p = end_poss_ptr[0]
            #for p in minus_tags[i_e:]:
            if p != pre_p:
                ret_p_ptr[0] = p
                ret_v_ptr[0] = max(pileup * scale_factor, baseline_value)
                ret_p_ptr += 1
                ret_v_ptr += 1
                I += 1
                pre_p = p
            pileup -= 1
            end_poss_ptr += 1

    ret_p.resize( I, refcheck=False )
    ret_v.resize( I, refcheck=False )

    return tmp

# quick pileup implemented in cython
cpdef list naive_quick_pileup ( np.ndarray[np.int32_t, ndim=1] sorted_poss, int extension):
    """Simple pileup, every tag will be extended left and right with length `extension`.

    Note: Assumption is that `poss` has to be pre-sorted! There is no
    check on whether it's sorted.
    """
    cdef:
        long i_s, i_e, i, I
        int a, b, p, pre_p, pileup
        long l = sorted_poss.shape[0]
        np.ndarray[np.int32_t, ndim=1] start_poss
        np.ndarray[np.int32_t, ndim=1] end_poss
        np.ndarray[np.int32_t, ndim=1] ret_p
        np.ndarray[np.float32_t, ndim=1] ret_v
        # pointers are used for numpy arrays
        int32_t * start_poss_ptr
        int32_t * end_poss_ptr
        int32_t * ret_p_ptr     # pointer for position array
        float32_t * ret_v_ptr     # pointer for value array

    start_poss = sorted_poss - extension
    start_poss[ start_poss < 0 ] = 0
    end_poss = sorted_poss + extension

    start_poss_ptr = <int32_t *> start_poss.data
    end_poss_ptr = <int32_t *> end_poss.data

    
    ret_p = np.zeros( 2*l, dtype="int32" )
    ret_v = np.zeros( 2*l, dtype="float32" )

    ret_p_ptr = <int32_t *> ret_p.data
    ret_v_ptr = <float32_t *> ret_v.data

    i_s = 0                         # index of plus_tags
    i_e = 0                         # index of minus_tags
    I = 0
    
    pileup = 0
    if l == 0: return ret
    pre_p = min(start_poss_ptr[0], end_poss_ptr[0])

    if pre_p != 0:
        # the first chunk of 0
        ret_p_ptr[0] = pre_p
        ret_v_ptr[0] = 0
        ret_p_ptr += 1
        ret_v_ptr += 1
        I += 1

    pre_v = pileup

    while i_s < l and i_e < l:
        if start_poss_ptr[0] < end_poss_ptr[0]:
            p = start_poss_ptr[0]
            if p != pre_p:
                ret_p_ptr[0] = p
                ret_v_ptr[0] = pileup
                ret_p_ptr += 1
                ret_v_ptr += 1
                I += 1
                pre_p = p
            pileup += 1
            i_s += 1
            start_poss_ptr += 1
        elif start_poss_ptr[0] > end_poss_ptr[0]:
            p = end_poss_ptr[0]
            if p != pre_p:
                ret_p_ptr[0] = p
                ret_v_ptr[0] = pileup
                ret_p_ptr += 1
                ret_v_ptr += 1
                I += 1
                pre_p = p
            pileup -= 1
            i_e += 1
            end_poss_ptr += 1
        else:
            i_s += 1
            i_e += 1
            start_poss_ptr += 1
            end_poss_ptr += 1

    # add rest of end positions
    if i_e < l:
        for i in range(i_e, l):
            p = end_poss_ptr[0]
            if p != pre_p:
                ret_p_ptr[0] = p
                ret_v_ptr[0] = pileup
                ret_p_ptr += 1
                ret_v_ptr += 1
                I += 1
                pre_p = p
            pileup -= 1
            end_poss_ptr += 1

    ret_p.resize( I, refcheck=False )
    ret_v.resize( I, refcheck=False )

    return [ ret_p, ret_v ]

# general function to compare two pv arrays in cython.
cpdef list over_two_pv_array ( list pv_array1, list pv_array2, func="max" ):
    """Merge two position-value arrays. For intersection regions, take
    the maximum value within region.

    pv_array1 and pv_array2 are [p,v] type lists, same as the output
    from quick_pileup function. 'p' and 'v' are numpy arrays of int32
    and float32.

    available operations are 'max', 'min', and 'mean'
    """

    cdef:
        int pre_p, p1, p2
        float v1, v2
        np.ndarray[np.int32_t, ndim=1] a1_pos, a2_pos, ret_pos
        np.ndarray[np.float32_t, ndim=1] a1_v, a2_v, ret_v
        int32_t * a1_pos_ptr
        int32_t * a2_pos_ptr
        int32_t * ret_pos_ptr
        float32_t * a1_v_ptr
        float32_t * a2_v_ptr
        float32_t * ret_v_ptr
        long l1, l2, l, i1, i2, I

    if func == "max":
        f = max
    elif func == "min":
        f = min
    elif func == "mean":
        f = mean
    else:
        raise Exception("Invalid function")

    [ a1_pos, a1_v ] = pv_array1
    [ a2_pos, a2_v ] = pv_array2
    ret_pos = np.zeros( a1_pos.shape[0] + a2_pos.shape[0], dtype="int32" )
    ret_v   = np.zeros( a1_pos.shape[0] + a2_pos.shape[0], dtype="float32" )

    a1_pos_ptr = <int32_t *> a1_pos.data
    a1_v_ptr = <float32_t *> a1_v.data
    a2_pos_ptr = <int32_t *> a2_pos.data
    a2_v_ptr = <float32_t *> a2_v.data
    ret_pos_ptr = <int32_t *> ret_pos.data
    ret_v_ptr = <float32_t *> ret_v.data

    l1 = a1_pos.shape[0]
    l2 = a2_pos.shape[0]

    i1 = 0
    i2 = 0
    I = 0

    pre_p = 0                   # remember the previous position in the new bedGraphTrackI object ret

    while i1 < l1 and i2 < l2:
        ret_v_ptr[0] =  f( a1_v_ptr[0], a2_v_ptr[0] )
        I += 1
        if a1_pos_ptr[0] < a2_pos_ptr[0]:
            # clip a region from pre_p to p1, then set pre_p as p1.
            ret_pos_ptr[0] = a1_pos_ptr[0]
            ret_pos_ptr += 1
            ret_v_ptr += 1
            pre_p = a1_pos_ptr[0]
            # call for the next p1 and v1
            a1_pos_ptr += 1
            a1_v_ptr += 1
            i1 += 1
        elif a1_pos_ptr[0] > a2_pos_ptr[0]:
            # clip a region from pre_p to p2, then set pre_p as p2.
            ret_pos_ptr[0] = a2_pos_ptr[0]
            ret_pos_ptr += 1
            ret_v_ptr += 1
            pre_p = a2_pos_ptr[0]
            # call for the next p1 and v1
            a2_pos_ptr += 1
            a2_v_ptr += 1
            i2 += 1
        else:
            # from pre_p to p1 or p2, then set pre_p as p1 or p2.
            ret_pos_ptr[0] = a1_pos_ptr[0]
            ret_pos_ptr += 1
            ret_v_ptr += 1
            pre_p = a1_pos_ptr[0]
            # call for the next p1, v1, p2, v2.
            a1_pos_ptr += 1
            a1_v_ptr += 1
            i1 += 1
            a2_pos_ptr += 1
            a2_v_ptr += 1
            i2 += 1

    ret_pos.resize( I, refcheck=False )
    ret_v.resize( I, refcheck=False )
    return [ret_pos, ret_v]

cpdef naive_call_peaks ( list pv_array, float min_v, float max_v = 1e30, int max_gap = 50, int min_length = 200 ):
    cdef:
        int peak_length, pre_p, p, i, summit, tstart, tend
        long x                  # index used for searching the first peak
        double v, summit_value, tvalue
        bytes chrom
        set chrs
        list peak_content       # (pre_p, p, v) for each region in the peak
        list ret_peaks = []     # returned peak summit and height

    peak_content = []
    ( ps,vs ) = pv_array
    psn = iter(ps).__next__         # assign the next function to a viable to speed up
    vsn = iter(vs).__next__
    x = 0
    pre_p = 0                   # remember previous position
    while True:
        # find the first region above min_v
        try:                    # try to read the first data range for this chrom
            p = psn()
            v = vsn()
        except:
            break
        x += 1                  # index for the next point
        if v > min_v:
            peak_content = [ ( pre_p , p, v ), ]
            pre_p = p
            break               # found the first range above min_v
        else:
            pre_p = p

    for i in range( x, len( ps ) ):
        # continue scan the rest regions
        p = psn()
        v = vsn()
        if v <= min_v: # not be detected as 'peak'
            pre_p = p
            continue
        # for points above min_v
        # if the gap is allowed
        # gap = pre_p - peak_content[-1][1] or the dist between pre_p and the last p
        if pre_p - peak_content[ -1 ][ 1 ] <= max_gap:
            peak_content.append( ( pre_p, p, v ) )
        else:
            # when the gap is not allowed, close this peak IF length is larger than min_length
            if peak_content[ -1 ][ 1 ] - peak_content[ 0 ][ 0 ] >= min_length:
                __close_peak( peak_content, ret_peaks, max_v, min_length )
            # reset and start a new peak
            peak_content = [ ( pre_p, p, v ), ]
        pre_p = p

    # save the last peak
    if peak_content:
        if peak_content[ -1 ][ 1 ] - peak_content[ 0 ][ 0 ] >= min_length:        
            __close_peak( peak_content, ret_peaks, max_v, min_length )
    return ret_peaks

cdef void __close_peak( peak_content, peaks, float max_v, int min_length ):
    """Internal function to find the summit and height

    If the height is larger than max_v, skip
    """
    tsummit = []
    summit = 0
    summit_value = 0
    for (tstart,tend,tvalue) in peak_content:
        if not summit_value or summit_value < tvalue:
            tsummit = [int((tend+tstart)/2),]
            summit_value = tvalue
        elif summit_value == tvalue:
            tsummit.append( int((tend+tstart)/2) )
    summit = tsummit[int((len(tsummit)+1)/2)-1 ]
    if summit_value < max_v:
        peaks.append( (summit, summit_value) )
    return



# cython: language_level=3
# cython: profile=True
# Time-stamp: <2020-12-03 11:46:58 Tao Liu>

"""Module Description:  IO Module for bedGraph file

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

# ------------------------------------
# python modules
# ------------------------------------
import io

from MACS3.Signal.BedGraph import bedGraphTrackI

# ------------------------------------
# constants
# ------------------------------------

# ------------------------------------
# C lib
# ------------------------------------

from libc.stdio cimport *
from libc.stdlib cimport *

# ------------------------------------
# Misc functions
# ------------------------------------

# ------------------------------------
# Classes
# ------------------------------------

cdef class bedGraphIO:
    """File Parser Class for bedGraph File.

    There are two assumptions in my bedGraphTrackI object:

    1. Continuous: the next region should be after the previous one
    unless they are on different chromosomes;

    2. Non-overlapping: the next region should never have overlaps
    with preceding region.

    If any of the above two criteria is violated, parsering will fail.
    """
    cdef:
        str bedGraph_filename

    def __init__ ( self, str bedGraph_filename ):
        """f must be a filename or a file handler.

        """
        self.bedGraph_filename = bedGraph_filename

    def build_bdgtrack (self, double baseline_value=0):
        """Use this function to return a bedGraphTrackI object.

        baseline_value is the value to fill in the regions not defined
        in bedGraph. For example, if the bedGraph is like:

        chr1  100 200  1
        chr1  250 350  2

        Then the region chr1:200..250 should be filled with
        baseline_value. Default of baseline_value is 0.
        """
        cdef bytes i

        data = bedGraphTrackI(baseline_value=baseline_value)
        add_func = data.add_loc
        # python open file
        bedGraph_file = open( self.bedGraph_filename, "rb" )

        for i in bedGraph_file:
            if i.startswith(b"track"):
                continue
            elif i.startswith(b"#"):
                continue
            elif i.startswith(b"browse"):
                continue
            else:
                fs = i.split()
                add_func(fs[0],atoi(fs[1]),atoi(fs[2]),atof(fs[3]))

        bedGraph_file.close()
        return data




# Time-stamp: <2020-11-30 13:49:34 Tao Liu>

"""Description: compare bdg files

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

import sys
import os

from MACS3.IO import BedGraphIO
from MACS3.Utilities.OptValidator import opt_validate_bdgcmp

from math import log as mlog

# ------------------------------------
# constants
# ------------------------------------

# ------------------------------------
# Misc functions
# ------------------------------------

# ------------------------------------
# Main function
# ------------------------------------

def run( options ):
    options = opt_validate_bdgcmp( options )
    info = options.info
    warn = options.warn
    debug = options.debug
    error = options.error
    
    scaling_factor = options.sfactor
    pseudo_depth = 1.0/scaling_factor   # not an actual depth, but its reciprocal, a trick to override SPMR while necessary.

    info("Read and build treatment bedGraph...")
    tbio = BedGraphIO.bedGraphIO(options.tfile)
    tbtrack = tbio.build_bdgtrack()

    info("Read and build control bedGraph...")
    cbio = BedGraphIO.bedGraphIO(options.cfile)
    cbtrack = cbio.build_bdgtrack()

    info("Build ScoreTrackII...")
    sbtrack = tbtrack.make_ScoreTrackII_for_macs( cbtrack, depth1 = pseudo_depth, depth2 = pseudo_depth )
    if abs(scaling_factor-1) > 1e-6:
        # Only for the case while your input is SPMR from MACS3 callpeak; Let's override SPMR.
        info("Values in your input bedGraph files will be multiplied by %f ..." % scaling_factor)
        sbtrack.change_normalization_method( ord('M') ) # a hack to override SPMR
    sbtrack.set_pseudocount( options.pseudocount )

    already_processed_method_list = []
    for (i, method) in enumerate(options.method):
        if method in already_processed_method_list:
            continue
        else:
            already_processed_method_list.append( method )

        info("Calculate scores comparing treatment and control by '%s'..." % method)
        if options.ofile:
            ofile = os.path.join( options.outdir, options.ofile[ i ] )
        else:
            ofile = os.path.join( options.outdir, options.oprefix + "_" + method + ".bdg" )
        # build score track
        if method == 'ppois':
            sbtrack.change_score_method( ord('p') )
        elif method == 'qpois':
            sbtrack.change_score_method( ord('q') )
        elif method == 'subtract':
            sbtrack.change_score_method( ord('d') )
        elif method == 'logFE':
            sbtrack.change_score_method( ord('f') )
        elif method == 'FE':
            sbtrack.change_score_method( ord('F') )
        elif method == 'logLR':             # log likelihood
            sbtrack.change_score_method( ord('l') )
        elif method == 'slogLR':             # log likelihood
            sbtrack.change_score_method( ord('s') )
        elif method == 'max':
            sbtrack.change_score_method( ord('M') )
        else:
            raise Exception("Can't reach here!")

        info("Write bedGraph of scores...")
        ofhd = open(ofile,"w")
        sbtrack.write_bedGraph(ofhd,name="%s_Scores" % (method.upper()),description="Scores calculated by %s" % (method.upper()), column = 3)
        info("Finished '%s'! Please check '%s'!" % (method, ofile))



# cython: language_level=3
# cython: profile=True
# Time-stamp: <2021-03-10 16:21:51 Tao Liu>

"""Module for SAPPER ReadAlignment class

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file COPYING included
with the distribution).
"""

# ------------------------------------
# python modules
# ------------------------------------
from cpython cimport bool

cdef extern from "stdlib.h":
    ctypedef unsigned int size_t
    size_t strlen(char *s)
    void *malloc(size_t size)
    void *calloc(size_t n, size_t size)
    void free(void *ptr)
    int strcmp(char *a, char *b)
    char * strcpy(char *a, char *b)
    long atol(char *bytes)
    int atoi(char *bytes)

# ------------------------------------
# constants
# ------------------------------------

__BAMDNACODE__ = b"=ACMGRSVTWYHKDBN"
__CIGARCODE__ = "MIDNSHP=X"
__DNACOMPLEMENT__ = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f !"#$%&\'()*+,-./0123456789:;<=>?@TBGDEFCHIJKLMNOPQRSAUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~\x7f\x80\x81\x82\x83\x84\x85\x86\x87\x88\x89\x8a\x8b\x8c\x8d\x8e\x8f\x90\x91\x92\x93\x94\x95\x96\x97\x98\x99\x9a\x9b\x9c\x9d\x9e\x9f\xa0\xa1\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9\xaa\xab\xac\xad\xae\xaf\xb0\xb1\xb2\xb3\xb4\xb5\xb6\xb7\xb8\xb9\xba\xbb\xbc\xbd\xbe\xbf\xc0\xc1\xc2\xc3\xc4\xc5\xc6\xc7\xc8\xc9\xca\xcb\xcc\xcd\xce\xcf\xd0\xd1\xd2\xd3\xd4\xd5\xd6\xd7\xd8\xd9\xda\xdb\xdc\xdd\xde\xdf\xe0\xe1\xe2\xe3\xe4\xe5\xe6\xe7\xe8\xe9\xea\xeb\xec\xed\xee\xef\xf0\xf1\xf2\xf3\xf4\xf5\xf6\xf7\xf8\xf9\xfa\xfb\xfc\xfd\xfe\xff' # A trans table to convert A to T, C to G, G to C, and T to A.

# -- CIGAR CODE --
#OP BAM  Description
#M  0    alignment match (can be a sequence match or mismatch) insertion to the reference
#I  1    insertion to the reference
#D  2    deletion from the reference
#N  3    skipped region from the reference
#S  4    soft clipping (clipped sequences present in SEQ)
#H  5    hard clipping (clipped sequences NOT present in SEQ)
#P  6    padding (silent deletion from padded reference)
#=  7    sequence match
#X  8    sequence mismatch
# -- -- -- -- -- --

# ------------------------------------
# Misc functions
# ------------------------------------

# ------------------------------------
# Classes
# ------------------------------------
cdef class ReadAlignment:
    cdef:
        bytes readname
        bytes chrom
        int lpos
        int rpos
        int strand              # strand information. 0 means forward strand, 1 means reverse strand.
        bytes binaryseq
        bytes binaryqual
        int l                   # length of read
        tuple cigar             # each item contains op_l|op
        bytes MD
        int n_edits             # number of edits; higher the number,
                                # more differences with reference
        bytes SEQ               # sequence of read regarding to + strand
        bytes QUAL              # quality of read regarding to + strand

    def __init__ ( self,
                   bytes readname,
                   bytes chrom, int lpos, int rpos,
                   int strand,
                   bytes binaryseq, 
                   bytes binaryqual,
                   tuple cigar,
                   bytes MD ):
        self.readname = readname
        self.chrom = chrom
        self.lpos = lpos
        self.rpos = rpos
        self.strand = strand
        self.binaryseq = binaryseq
        self.binaryqual = binaryqual
        self.l = len( binaryqual )
        self.cigar = cigar
        self.MD = MD
        self.n_edits = self.get_n_edits()
        (self.SEQ, self.QUAL) = self.__get_SEQ_QUAL()

    cdef int get_n_edits( self ):
        """The number is from self.cigar and self.MD.

        """
        cdef:
            int n_edits
            int i, cigar_op, cigar_op_l
            char c
        
        n_edits = 0
        for i in self.cigar:    # only count insertion or softclip
            cigar_op = i & 15
            cigar_op_l = i >> 4
            if cigar_op in [ 1, 4 ]:    # count Insertion or Softclip
                n_edits += cigar_op_l
        
        for c in self.MD:
            if (c > 64 and c < 91): # either deletion in query or mismatch
                n_edits += 1
        return n_edits

    def __str__ ( self ):
        c = self.chrom.decode()
        n = self.readname.decode()
        if self.strand:
            s = "-"
        else:
            s = "+"
        return f"{c}\t{self.lpos}\t{self.rpos}\t{n}\t{self.l}\t{s}"
    
    def __getitem__ ( self, keyname ):
        if keyname == "readname":
            return self.readname
        elif keyname == "chrom":
            return self.chrom
        elif keyname == "lpos":
            return self.lpos
        elif keyname == "rpos":
            return self.rpos
        elif keyname == "strand":
            return self.strand
        elif keyname == "SEQ":
            return self.SEQ
        elif keyname == "QUAL":
            return self.QUAL
        elif keyname == "n_edits":
            return self.n_edits
        elif keyname == "binaryseq":
            return self.binaryseq
        elif keyname == "binaryqual":
            return self.binaryqual
        elif keyname == "l":
            return self.l
        elif keyname == "cigar":
            return self.cigar
        elif keyname == "MD":
            return self.MD
        else:
            raise KeyError("No such key", keyname)

    def __getstate__ ( self ):
        return ( self.readname, self.chrom, self.lpos, self.rpos, self.strand, self.binaryseq, self.binaryqual, self.l, self.cigar, self.MD, self.n_edits, self.SEQ, self.QUAL )

    def __setstate__ ( self, state ):
        ( self.readname, self.chrom, self.lpos, self.rpos, self.strand, self.binaryseq, self.binaryqual, self.l, self.cigar, self.MD, self.n_edits, self.SEQ, self.QUAL ) = state

    # cpdef bytearray get_SEQ ( self ):
    #     """Convert binary seq to ascii seq.

    #     Rule: for each byte, 1st base in the highest 4bit; 2nd in the lowest 4bit. "=ACMGRSVTWYHKDBN" -> [0,15]

    #     Note: In BAM, if a sequence is mapped to reverse strand, the
    #     reverse complement seq is written in SEQ field. So the return
    #     value of this function will not be the original one if the
    #     read is mapped to - strand.
    #     """
    #     cdef:
    #         char c
    #         bytearray seq

    #     seq = bytearray(b"")
    #     for c in self.binaryseq:
    #         # high
    #         seq.append( __BAMDNACODE__[c >> 4 & 15] )
    #         # low
    #         seq.append( __BAMDNACODE__[c & 15] )
    #     if seq[-1] == b"=":
    #         # trim the last '=' if it exists
    #         seq = seq[:-1]
    #     return seq

    cdef tuple __get_SEQ_QUAL ( self ):
        """Get tuple of (SEQ, QUAL).

        Rule: for each byte, 1st base in the highest 4bit; 2nd in the lowest 4bit. "=ACMGRSVTWYHKDBN" -> [0,15]

        Note: In BAM, if a sequence is mapped to reverse strand, the
        reverse complement seq is written in SEQ field. So the return
        value of this function will not be the original one if the
        read is mapped to - strand. If you need to original one, do
        reversecomp for SEQ and reverse QUAL.
        """
        cdef:
            int i
            char c
            bytearray seq
            bytearray qual

        seq = bytearray(b"")
        qual = bytearray(b"")

        for i in range( len(self.binaryseq) ):
            c = self.binaryseq[ i ]
            # high
            seq.append( __BAMDNACODE__[c >> 4 & 15] )
            # low
            seq.append( __BAMDNACODE__[c & 15] )

        for i in range( len( self.binaryqual ) ):
            # qual is the -10log10 p or phred score. 
            qual.append( self.binaryqual[i] )
            
        if seq[-1] == b"=":
            # trim the last '=' if it exists
            seq = seq[:-1]
        assert len( seq ) == len( qual ), Exception("Lengths of seq and qual are not consistent!")

        # Example on how to get original SEQ and QUAL:
        #if self.strand:
        #    seq.reverse()
        #    #compliment
        #    seq = seq.translate( __DNACOMPLEMENT__ )
        #    qual.reverse()

        return ( bytes(seq), bytes(qual) )

    
    cpdef bytes get_FASTQ ( self ):
        """Get FASTQ format text.

        """
        cdef:
            bytes seq
            bytearray qual

        seq = self.SEQ
        qual = bytearray(self.QUAL)

        for i in range( len( self.QUAL ) ):
            # qual is the -10log10 p or phred score, to make FASTQ, we have to add 33
            qual[ i ] += 33
        
        # reverse while necessary
        if self.strand:
            seq = self.SEQ[::-1]
            #compliment
            seq = seq.translate( __DNACOMPLEMENT__ )
            qual = qual[::-1]
        else:
            seq = self.SEQ

        return b"@" + self.readname + b"\n" + seq + b"\n+\n" + qual + b"\n"

    cpdef bytearray get_REFSEQ ( self ):
        """Fetch reference sequence, using self.MD and self.cigar
        """
        cdef:
            char c
            bytearray seq, refseq
            int i, cigar_op, cigar_op_l
            bytearray MD_op
            int ind
            bool flag_del       # flag for deletion event in query

        seq = bytearray(self.SEQ)    # we start with read seq then make modifications

        # 2-step proces
        # First step: use CIGAR to edit SEQ to remove S (softclip) and I (insert)
        # __CIGARCODE__ = "MIDNSHP=X"
        # let ind be the index in SEQ
        ind = 0
        for i in self.cigar:
            cigar_op = i & 15
            cigar_op_l = i >> 4
            if cigar_op in [2, 5, 6]:     # do nothing for Deletion (we will
                                          # put sequence back in step 2),
                                          # Hardclip and Padding
                pass
            elif cigar_op in [0, 7, 8]:   # M = X alignment match (match or
                                          # mismatch)
                # do nothing and move ind
                ind += cigar_op_l
            elif cigar_op in [ 1, 4 ]:    # Remove for Insertion or Softclip
                seq[ ind : ind + cigar_op_l ] = b''

        # now the seq should be at the same length as rpos-lpos 

        # Second step: use MD string to edit SEQ to put back 'deleted
        # seqs' and modify mismatches

        # let ind be the index in SEQ again, from 0
        ind = 0
        MD_op = bytearray(b'')
        flag_del = False
        for c in self.MD:
            if c < 58 and c > 47:
                # means Match
                flag_del = False
                MD_op.append(c)
            elif (c > 64 and c < 91) and not flag_del:
                # An alphabet means Mismatch, Note, if MD is made
                # right, a mismatch should only be 1 letter surrounded
                # by digits.
                ind += int(MD_op)
                seq[ ind ] = c
                ind += 1
                # reset MD_op
                MD_op = bytearray(b'')
            elif (c > 64 and c < 91) and flag_del:
                seq[ ind:ind ] = [c,]
                ind += 1
            elif c == 94:
                # means Deletion in query. Now, insert a sequnce into
                # SEQ
                flag_del = True
                ind += int(MD_op)
                # reset MD_op
                MD_op = bytearray(b'')
            else:
                raise Exception("Don't understand this operator in MD: %c" % c)
            #print( seq.decode() )

        return seq
    
    cpdef get_base_by_ref_pos ( self, long ref_pos ):
        """Get base by ref position.

        """
        cdef:
           int relative_pos, p
        assert self.lpos <= ref_pos and self.rpos > ref_pos, Exception("Given position out of alignment location")
        relative_pos = ref_pos - self.lpos
        p = self.relative_ref_pos_to_relative_query_pos( relative_pos ) 

        if p == -1:             # located in a region deleted in query
            return None
        else:
            return __BAMDNACODE__[ (self.binaryseq[p//2] >> ((1-p%2)*4) ) & 15 ]

    cpdef get_bq_by_ref_pos ( self, long ref_pos ):
        """Get base quality by ref position. Base quality is in Phred scale.

        Returned value is the raw Phred-scaled base quality.

        """
        cdef:
           int relative_pos, p
        assert self.lpos <= ref_pos and self.rpos > ref_pos, Exception("Given position out of alignment location")
        relative_pos = ref_pos - self.lpos
        p = self.relative_ref_pos_to_relative_query_pos( relative_pos ) 

        if p == -1:             # located in a region deleted in query
            return None
        else:
            return self.binaryqual[p]

    cpdef tuple get_base_bq_by_ref_pos ( self, long ref_pos ):
        """Get base and base quality by ref position. Base quality is in Phred scale.

        Returned bq is the raw Phred-scaled base quality.
        """
        cdef:
           int relative_pos, p
        assert self.lpos <= ref_pos and self.rpos > ref_pos, Exception("Given position out of alignment location")
        relative_pos = ref_pos - self.lpos
        p = self.relative_ref_pos_to_relative_query_pos( relative_pos ) 

        if p == -1:             # located in a region deleted in query
            return None
        else:
            return ( __BAMDNACODE__[ (self.binaryseq[p//2] >> ((1-p%2)*4) ) & 15 ], self.binaryqual[p] )

    cpdef tuple get_variant_bq_by_ref_pos ( self, long ref_pos ):
        """Get any variants (different with reference) and base quality by ref position. 

        variants will be 

        1) =, if identical

        2) A/T/C/G, if SNV

        3) -, if the reference base is deleted, in this case, bq will
        be the highest possible bq, which is 93.

        4) ^<A/T/C/G>+, if there is an insertion at the location

        Base quality is the raw Phred-scaled base quality.

        """
        cdef:
           int i, m, n
           int res, p, op, op_l
           int pos
           bool tip
           bytearray refseq
           bytes p_refseq, p_seq
           bytearray seq_array
           bytearray bq_array

        assert self.lpos <= ref_pos and self.rpos > ref_pos, Exception("Given position out of alignment location")

        res = ref_pos - self.lpos          # residue
        p = 0
        
        refseq = self.get_REFSEQ()
        p_refseq =  refseq[ res ]
        # -- CIGAR CODE --
        #OP BAM  Description
        #M  0    alignment match (can be a sequence match or mismatch) insertion to the reference
        #I  1    insertion to the reference
        #D  2    deletion from the reference
        #N  3    skipped region from the reference
        #S  4    soft clipping (clipped sequences present in SEQ)
        #H  5    hard clipping (clipped sequences NOT present in SEQ)
        #P  6    padding (silent deletion from padded reference)
        #=  7    sequence match
        #X  8    sequence mismatch

        seq_array = bytearray( b'' )
        bq_array = bytearray( b'' )

        for m in range( len(self.cigar) ):
            i = self.cigar[ m ]
            op = i & 15
            op_l = i >> 4
            if op in [0, 7, 8]:         # M = X alignment match (match or mismatch)
                if res < op_l - 1:
                    # in the range of a CIGAR operator
                    p += res
                    # find the position, now get the ref
                    pos = p
                    seq_array.append( __BAMDNACODE__[ (self.binaryseq[ p//2 ] >> ( (1-p%2)*4 ) ) & 15 ] )
                    bq_array.append( self.binaryqual[ p ] )
                    break
                elif res == op_l - 1:
                    p += res
                    pos = p
                    seq_array.append( __BAMDNACODE__[ (self.binaryseq[ p//2 ] >> ( (1-p%2)*4 ) ) & 15 ] )
                    bq_array.append( self.binaryqual[ p ] )
                    # now add any insertion later on
                    # get next cigar
                    if m + 1 == len( self.cigar ):
                        break
                    i = self.cigar[ m + 1 ]
                    op = i & 15
                    op_l = i >> 4
                    if op == 1:           #insertion
                        for n in range( op_l ):
                            p += 1
                            seq_array.append( __BAMDNACODE__[ (self.binaryseq[ p//2 ] >> ( (1-p%2)*4 ) ) & 15 ] )
                            bq_array.append( self.binaryqual[ p ] )
                        #print self.SEQ, seq_array
                    break
                else:
                    # go to the next cigar code
                    p += op_l
                    res -= op_l
            elif op in [ 2, 3 ]: # D N
                if res < op_l:
                    # find the position, however ...
                    # position located in a region in reference that not exists in query
                    pos = p
                    seq_array.append( b'*' )
                    bq_array.append( 93 )   #assign 93 for deletion
                    break
                else:
                    # go to the next cigar code
                    res -= op_l
            elif op == 1 :      # Insertion
                p += op_l
                # if res == 0:    # no residue left, so return a chunk of inserted sequence
                #     print "shouldn't run this code"
                #     # first, add the insertion point
                #     seq_array = bytearray( b'~' )
                #     bq_array.append( self.binaryqual[ p ] )
                #     # then add the inserted seq
                #     for i in range( op_l ):
                #         p += 1
                #         seq_array.append( __BAMDNACODE__[ (self.binaryseq[ p//2 ] >> ( (1-p%2)*4 ) ) & 15 ]  ) 
                #         bq_array.append( self.binaryqual[ p ] )
                #     break
                # else:
                #     p += op_l
            elif op == 4 :      # Softclip. If it's Softclip, we'd better not return the extra seq
                p += op_l

        if pos == 0 or pos == self.l - 1:
            tip = True
        else:
            tip = False
                
        return ( seq_array, bq_array, self.strand, tip, pos )
        # last position ?
        #raise Exception("Not expected to see this")

    cdef int relative_ref_pos_to_relative_query_pos ( self, long relative_ref_pos ):
        """Convert relative pos on ref to pos on query.
        """
        cdef:
            int p, res, op, op_l
        p = 0
        res = relative_ref_pos
        
        for i in self.cigar:
            op = i & 15
            op_l = i >> 4
            if op in [0, 7, 8]:         # M = X alignment match (match or mismatch)
                if res < op_l:
                    p += res
                    return p
                else:
                    p += op_l
                    res -= op_l
            elif op in [ 2, 3 ]: # D N
                if res < op_l:
                    # position located in a region in reference that not exists in query
                    return -1
                else:
                    res -= op_l
            elif op in [ 1, 4 ]:       # I
                p += op_l
        return p


### End ###

